import base64
import json

from hashlib import sha256
from cryptography.hazmat.primitives import hashes
from cryptography.hazmat.primitives.asymmetric import ec
from cryptography.hazmat.primitives import serialization

class BCct:
  SIGN      = 'EE_SIGN'
  SENDER    = 'EE_SENDER'
  HASH      = 'EE_HASH'
  ADDR_PREFIX   = "0xai_"

NON_DATA_FIELDS = [BCct.HASH, BCct.SIGN, BCct.SENDER]


def create_private_key():
  sk = ec.generate_private_key(curve=ec.SECP256K1())
  return sk


def pk_to_addr(pk):
  data = pk.public_bytes(
    encoding=serialization.Encoding.X962,
    format=serialization.PublicFormat.CompressedPoint,
  )
  txt = BCct.ADDR_PREFIX + base64.urlsafe_b64encode(data).decode()
  return txt

def addr_to_pk(addr):
  simple_addr = addr.replace(BCct.ADDR_PREFIX, '')
  bpublic_key = base64.urlsafe_b64decode(simple_addr)
  pk = ec.EllipticCurvePublicKey.from_encoded_point(
    curve=ec.SECP256K1(), 
    data=bpublic_key
  )
  return pk
  

def sign_object(obj, sk):
  simple_object = {k : obj[k] for k in obj if k not in NON_DATA_FIELDS}
  data = json.dumps(simple_object, sort_keys=True, separators=(',',':')).encode()
  str_digest = sha256(data).hexdigest()
  bin_digest = str_digest.encode()
  signature = sk.sign(
    data=bin_digest,
    signature_algorithm=ec.ECDSA(hashes.SHA256())
    )
  str_signature = base64.urlsafe_b64encode(signature).decode()
  obj[BCct.SENDER] = pk_to_addr(sk.public_key())
  obj[BCct.SIGN] = str_signature
  obj[BCct.HASH] = str_digest
  return str_signature


def verify_signature(obj):
  addr = obj.get(BCct.SENDER)
  str_sign = obj.get(BCct.SIGN)
  valid = False
  assert addr is not None and str_sign is not None
  simple_object = {k : obj[k] for k in obj if k not in NON_DATA_FIELDS}
  data = json.dumps(simple_object, sort_keys=True, separators=(',',':')).encode()
  str_digest = sha256(data).hexdigest()
  bin_digest = str_digest.encode()
  pk = addr_to_pk(addr)
  try:
    assert str_digest == obj[BCct.HASH], "DIGEST_ERROR"
    signature = base64.urlsafe_b64decode(str_sign)
    pk.verify(signature, bin_digest, ec.ECDSA(hashes.SHA256()))
    valid = True
  except Exception as exc:
    error_msg = str(exc)
    if len(error_msg) == 0:
      error_msg = exc.__class__.__name__
    print(error_msg)
  return valid
  

if __name__ == '__main__':
  # testing code
  d = {'9'  : 9, '2':2, '3':3, '10':{'2':2,'100':100, '1':1}}
  sk = create_private_key()
  s = sign_object(d, sk)
  message = json.dumps(d)
  print("\n{}\n---------------".format(json.dumps(d, indent=4)))
  d1 = json.loads(message)
  d2 = json.loads(message)
  d0 = json.loads(message)
  d1['9'] = None
  d2[BCct.SIGN] = 'aaaabbbbcccc' # faking the signature
  print("Bandit 1:")
  v = verify_signature(d1)
  print(v)
  print("---------------")
  print("Bandit 2:")
  v = verify_signature(d2)
  print(v)
  print("---------------")
  print("Good guy:")
  v = verify_signature(d0)
  print(v)
  print("---------------")
  
  # '{"9": 9, "2": 2, "3": 3, "10": {"2": 2, "1": 1}, "EE_SENDER": "0xai_AnYCe3QIJiVWS-borGH4S0_Hv9AKM5ybP2tCWkH3zFJE", "EE_SIGN": "MEQCICp9p9MO-urtQXEJ8rrE7Kwa930PP86GElkE16YFtSWQAiA_0HChXFIzZIS0wQ0Ut2HsNuz_ZO90LHA019h1S4IDYA==", "EE_HASH": "c49af6451a698441f212539ac9b9d196bbf5a893ef8dfc430f21db967ec7263d"}'
  external_message = "put here your json" 
  if external_message != "put here your json":
    external_data = json.loads(external_message)
    v = verify_signature(external_data)
    print("Result on external message: {}".format(v))
  
  
  


# global dependencies
import gc
import random
import torch as th
import pandas as pd
import numpy as np
import cv2
import os
import traceback

from time import sleep
from collections import OrderedDict

# local dependencies
from naeural_core import DecentrAIObject
from decentra_vision.draw_utils import DrawUtils
from naeural_core.serving.serving_manager import ServingManager
from plugins.serving.model_testing.utils import Dataset
from plugins.serving.model_testing.utils import ServingUtils


class Base(DecentrAIObject):
  def __init__(self,
               model_name,
               test_files,
               sleep_time=10,
               save_plots=True,
               show_plots=False,
               label_extension='json',
               nr_warmup=2,
               nr_predicts=20,
               inprocess=False,
               serving_manager=None,
               print_errors=True,
               subdir_keys=None,
               **kwargs):
    """

    :param model_name: tuple(CLASS, NAME)
      Passed to ServingManager.start_server server_name

    :param test_files: dict(dataset_name: [list of paths])
      Files to run inference on

    :param sleep_time: int
      Sleep time at startup

    :param save_plots: bool
      Whether to save ploted images to disk

    :param show_plots: bool
      Whether to show plotted images

    :param nr_warmup: int
      How many warmup steps to run

    :param nr_predicts: int
      How many predicts to run in order to get mean resource consumption

    :param inprocess: bool
      Whether to run inprocess serving processes

    :param serving_manager: ServingManager object

    :param print_errors: bool
      Whether to display inference errors

    :param subdir_keys: list
      keys used to better organize the saved results (mainly plots) in subdirectories

    :param kwargs:
    """
    super().__init__(**kwargs)
    self._model_name = model_name
    self._test_files = test_files
    self._sleep_time = sleep_time
    self._save_plots = save_plots
    self._show_plots = show_plots
    # self._gpu_device = gpu_device
    self._serving_manager = serving_manager
    # if self._gpu_device is None:
    #   self._gpu_device = 1 if th.cuda.device_count() > 1 else 0
    self._nr_predicts = nr_predicts
    self._nr_warmup = nr_warmup
    self._inprocess = inprocess
    self._print_errors = print_errors
    self.subdir_keys = [] if subdir_keys is None else subdir_keys
    self.current_dct_params = {}

    if self._inprocess is True:
      self.P("WARNING !!! 'inprocess' is set to True which may yield invalid values because of the inability to do a complete cleanup", color='r')
    #
    # self._lst_imgs = None
    # self._lst_img_names = None

    self._dct_res = {
      'MODEL': [],
      'TIME': [],
      'TOTAL_MEM': [],
      'INITIALIZATION_MEM': [],
      'TIME_PER_IMAGE': [],
      'NUMBER_OF_IMAGES': [],
      'SUCCESS': []
    }
    self._test_datasets = {}
    self._test_datasets_paths = {}
    self.label_extension = label_extension
    self._preds = OrderedDict()
    self._init()
    return

  def set_loop_stage(*args, **kwargs):
    return

  def _init(self):
    """Starting serving manager, starting utils. Loading data."""

    if self._serving_manager is None:
      self._serving_manager = ServingManager(
        shmem={},
        log=self.log,
        prefix_log='[TFMGR]',
        owner=self,
      )
    self._painter = DrawUtils(log=self.log)
    self._dataset = Dataset(log=self.log)
    self._serving_utils = ServingUtils(log=self.log)
    self._load_data()
    return

  def _load_data(self):
    """Loading testing dataset from indicated list of files"""

    self.log.P("Preparing data...", color='r')
    self.load_data()
    self.log.P("Done preparing data.")
    return

  def load_data(self):
    for dataset_name, dataset_paths in self._test_files.items():
      self._test_datasets[dataset_name] = self._dataset.load_images(dataset_paths)
      self._test_datasets_paths[dataset_name] = [
        f'{os.path.splitext(dataset_path)[0]}.{self.label_extension}'
        for dataset_path in dataset_paths
      ]
    # self._lst_imgs = list(dct.values())
    # self._lst_img_names = list(dct.keys())
    return

  def run(self, dct_test={}, dct_params={}, _lst_images=None, kill_serving_manager=True):
    """
    TODO: big ambiguity in dct_test vs dct_params

    Execute inference testing

    Video file testing params:
      _lst_images = force the use of the given _lst_images
      kill_serving_manager = set to false in video file testing in order to not create a new manager for every movie batch
    """
    gpu_id = 0
    preds = None

    gpu = dct_test.get('DEFAULT_DEVICE', None)

    if "MAX_BATCH_FIRST_STAGE" not in dct_test:
      self.P("WARNING! Parameter 'MAX_BATCH_FIRST_STAGE' not set. Using default value 5", color='r')

    batch_first_stage = dct_test.get("MAX_BATCH_FIRST_STAGE", 5)
    batch_second_stage = dct_test.get("MAX_BATCH_SECOND_STAGE", 1)

    batch_size = max(batch_first_stage, batch_second_stage)

    run_dct = {
      'DEFAULT_DEVICE': None,
      'MODEL': None,
      'TIME': None,
      'TOTAL_MEM': None,
      'INITIALIZATION_MEM': None,
      'TIME_PER_IMAGE': None,
      'NUMBER_OF_IMAGES': None,
      'SUCCESS': None,
      'PREDICTS': self._nr_predicts,
      'WARMUPS': self._nr_warmup,
      **dct_test
    }
    try:
      # TODO Bleo: Implement cpu testing
      if gpu is not None:
        gpus_info = self.log.gpu_info(show=True, mb=True)[gpu_id]
        gpu_id = [i for i, gpu_info in enumerate(gpus_info) if gpus_info['NAME'] == th.cuda.get_device_name(gpu)][0]

      self.log.P("Running test for {}".format(dct_test), color='g')
      g1 = self.log.gpu_info(show=True, mb=True)[gpu_id]
      self.log.P("Free memory before start: {}".format(g1['FREE_MEM']))

      upstream_config = {
        **dct_test,
        **dct_params
      }
      str_model_name = self._serving_manager.start_server(
        server_name=self._model_name,
        inprocess=self._inprocess,
        upstream_config=upstream_config,
      )
      run_dct['MODEL'] = str_model_name

      g15 = self.log.gpu_info(show=True, mb=True)[gpu_id]
      initialzation_mem = g1['FREE_MEM'] - g15['FREE_MEM']
      run_dct['INITIALIZATION_MEM'] = initialzation_mem

      self.log.P("Model is ready. Check nvidia-smi in next {} seconds...".format(self._sleep_time))
      sleep(self._sleep_time)

      self.log.P("Running warm-up predict...")
      if _lst_images is None:
        _lst_images = self.get_images(dct_test['dataset_name'])

      no_imgs = len(_lst_images)
      _lst_images = (_lst_images * np.ceil(batch_size / no_imgs).astype(int))[:max(batch_size, no_imgs)]
      run_dct['NUMBER_OF_IMAGES'] = len(_lst_images)

      yolo_model_inputs = self._serving_utils.get_model_input(_lst_images)
      for _ in range(self._nr_warmup):
        preds = self._serving_manager.predict(self._model_name, yolo_model_inputs)
      # endfor

      if preds is None and self._nr_warmup > 0:
        raise Exception("Error raised in predict")

      self.log.reset_timers()

      self.log.P("Running batch predict...")
      yolo_model_inputs = self._serving_utils.get_model_input(_lst_images)
      preds = self._serving_manager.predict(self._model_name, yolo_model_inputs)

      if preds is None:
        raise Exception("Error raised in predict")

      for _ in range(self._nr_predicts - 1):
        random.shuffle(_lst_images)
        yolo_model_inputs = self._serving_utils.get_model_input(_lst_images)
        self._serving_manager.predict(self._model_name, yolo_model_inputs)
      # endfor

      self._preds[len(self._preds)] = preds

      g2 = self.log.gpu_info(show=True, mb=True)[gpu_id]
      self.log.P("Free memory after run: {}".format(g2['FREE_MEM']))
      total_model_mem = g1['FREE_MEM'] - g2['FREE_MEM']
      run_dct['TOTAL_MEM'] = total_model_mem
      self.log.P("Model memory {:.0f} MB".format(total_model_mem), color='m')

      self.log.P("Showing results & timers...", color='g')
      self.log.show_timers(title="on {} for '{}' with {}".format(
          self.log.get_machine_name(),
          self._model_name, dct_test
        ),
        div=max(len(_lst_images), 1)
      )

      if dct_test.get('dataset_name') is not None:
        self.plot(**{**dct_test, **dct_params})
        res = self.score(**{**dct_test, **dct_params})
        if res is not None:
          run_dct['SCORE'] = res
        # endif res None
      # endif dataset available

      if not self._show_plots and not self._save_plots:
        self.log.P("Skipped showing images & inferences.")
      # endif

      g3 = self.log.gpu_info(show=True, mb=True)[gpu_id]
      self.log.P("Free memory after shutdown: {}".format(g3['FREE_MEM']))

      if kill_serving_manager:
        self._serving_manager.stop_server(server_name=self._model_name)

      timer_prefix = 'local_pred_' if self._inprocess else 'remote_pred_'
      run_dct['TIME'] = self.log.get_timer(timer_prefix + str_model_name.upper())['MEAN']
      run_dct['TIME_PER_IMAGE'] = self.log.get_timer(timer_prefix + str_model_name.upper())['MEAN'] / len(_lst_images)
      run_dct['SUCCESS'] = True
    except:
      run_dct['SUCCESS'] = False
      if self._print_errors:

        info = traceback.format_exc()
        self.log.P("Inference error: {}".format(info), color='r')
    # endtry

    for k, v in run_dct.items():
      if k not in self._dct_res:
        self._dct_res[k] = []
      self._dct_res[k].append(v)

    gc.collect()
    th.cuda.empty_cache()
    return preds

  def get_results(self):
    df = pd.DataFrame(self._dct_res)
    return df

  def show_results(self):
    df = self.get_results()
    self.log.set_nice_prints()
    self.log.P("Results on {}:\n{}".format(self.log.get_machine_name(), df))
    return

  @property
  def _testing_subfolder_path(self):
    return self._testing_subfolder_path_helper()

  def _testing_subfolder_path_helper(self, skip_subdirs=False):
    res = os.path.join('testing', self.log.file_prefix, self.__name__, self._model_name)
    if not skip_subdirs and len(self.subdir_keys) > 0:
      values = [self.current_dct_params.get(key) for key in self.subdir_keys if key in self.current_dct_params]
      str_values = [str(value) for value in values]
      res = os.path.join(res, *str_values)
    # endif subdir_keys
    return res

  def save_results(self):
    df = self.get_results()
    df_name = 'results.csv'
    self.log.save_dataframe(df, df_name, folder='output',
                            subfolder_path=self._testing_subfolder_path_helper(skip_subdirs=True))

  def run_tests(self, lst_tests, dct_params, save_results=True):
    # TODO: big ambiguity between lists of test dicts and dct_params
    for dataset_name in self.get_dataset_names():
      for dct_test in lst_tests:
        test_params = {'dataset_name': dataset_name, **dct_test}
        self.current_dct_params = test_params
        self.run(test_params, dct_params)
    # endfor
    self.show_results()
    if save_results:
      self.save_results()
    return self.get_results()

  def plot(self, dataset_name, **kwargs):
    """Draw plots specific to each kind of model"""

    raise NotImplementedError()
    return

  def score(self, dataset_name, **kwargs):
    """Score computing for specific model"""

    return None

  def plot_video(self, cv2_video_writer, lst_imgs, preds, **kwargs):
    """Based on a lst of imgs and preds save a video"""
    raise NotImplementedError()
    return

  def get_images(self, dataset_name):
    return self._test_datasets[dataset_name]

  def get_dataset_names(self):
    return list(self._test_datasets.keys())

  def get_dataset_labels_path(self, dataset_name):
    return self._test_datasets_paths[dataset_name]

  def get_preds(self, idx=None):
    preds = self._preds
    if idx is not None:
      preds = self._preds.get(idx, None)
    return preds

  def get_last_preds(self):
    dct_preds = self.get_preds()
    keys = list(dct_preds.keys())
    preds = dct_preds[keys[-1]]
    return preds

  def get_model_name(self):
    return self._model_name

  def run_video_test(self, video_fn, test_params={}, flip_video=False, movie_batch_size=50, plot_video_kwargs={}):
    """

    :param video_fn:
    :param flip_video:
      False - nothing happens
      Rotate code - passed to cv2.rotate
    :param movie_batch_size:
    :return:
    """
    cap = cv2.VideoCapture(video_fn)
    while not cap.isOpened():
      cap = cv2.VideoCapture(video_fn)

    frames = []
    fps = cap.get(cv2.CAP_PROP_FPS)
    pos_frame = cap.get(cv2.CAP_PROP_POS_FRAMES)
    while True:
      flag, frame = cap.read()
      if flip_video is not False:
        frame = cv2.rotate(frame, flip_video)
      frame_rgb = frame[:, :, ::-1]
      frame_rgb = np.ascontiguousarray(frame_rgb)
      frames.append(frame_rgb)

      if cap.get(cv2.CAP_PROP_POS_FRAMES) == cap.get(cv2.CAP_PROP_FRAME_COUNT):
          # If the number of captured frames is equal to the total number of frames,
          # we stop
        break

    output_video_file = '{}_{}'.format(self.log.file_prefix, os.path.split(video_fn)[-1])

    output_video_folder = os.path.join(
      self.log.get_output_folder(),
      'video_testing',
    )
    if not os.path.exists(output_video_folder):
      os.makedirs(output_video_folder)

    output_video_fn = os.path.join(output_video_folder, output_video_file)

    cv2_video_writer = cv2.VideoWriter(
      output_video_fn,
      cv2.VideoWriter_fourcc(*"XVID"),
      fps,
      (frames[0].shape[1], frames[0].shape[0])
    )

    for batch_id in range(len(frames) // movie_batch_size):
      self.P("Processing movie batch {}/{}".format(batch_id + 1, len(frames) // movie_batch_size))
      batch = frames[batch_id * movie_batch_size: (batch_id + 1) * movie_batch_size]
      batch_results = self.run(dct_test={**test_params, "MAX_BATCH_FIRST_STAGE": 5},
                               _lst_images=batch, kill_serving_manager=False)
      self.plot_video(
        cv2_video_writer=cv2_video_writer,
        lst_imgs=batch,
        preds=batch_results,
        **plot_video_kwargs
      )
    cv2_video_writer.release()

    a = 0


if __name__ == '__main__':
  from naeural_core import Logger
  from plugins.serving.testing.th_yolov5l6s.data import files as TEST_FILES

  log = Logger('MPTF', base_folder='.', app_folder='_solisbox', TF_KERAS=False)

  # defining script constants
  GPU_DEVICE = 1 if th.cuda.device_count() > 1 else 0
  TESTS = [
    # {
    #   'USE_AMP'         : False,
    #   'USE_FP16'        : False,
    #   'GPU_PREPROCESS'  : False,
    # },

    {
      'USE_AMP': True,
      'USE_FP16': False,
      'GPU_PREPROCESS': True,
      'DEFAULT_DEVICE': 'cuda:' + str(GPU_DEVICE)
    },

    # {
    #   'USE_AMP'         : True,
    #   'USE_FP16'        : False,
    #   'GPU_PREPROCESS'  : True,
    # },

    # {
    #   'USE_AMP'         : False,
    #   'USE_FP16'        : True,
    #   'GPU_PREPROCESS'  : True,
    # },
  ]
  COVERED = 'eff_det0'
  MODEL_NAME = ('th_y5l6s', 'default')

  test = Base(
    log=log,
    model_name=MODEL_NAME,
    test_files=TEST_FILES,
    gpu_device=GPU_DEVICE,
  )

  dct_params = {
    "NEW_PARAM": 0,
    "COVERED_SERVERS": ['th_y5l6', COVERED],
    "NMS_CONF_THR": 0.05,
    "WARMUP_BATCH": len(test.get_images())
  }
  test.run_tests(
    lst_tests=TESTS,
    dct_params=dct_params
  )



from naeural_core import constants as ct
import math
import random


SIZE = 3
BABY_STATE = 0
ADULT_STATE = 1
RETIRED_STATE = 2
BABY_STEPS_LIMIT = 5
BACK_TRIES_LIMIT = 1000
LIFE_STATES = ['BABY', 'ADULT', 'RETIRED']
EPSILON = 0.0001

TYPE_SNAKE = 'SNAKE'
TYPE_HEDGEHOG = 'HEDGEHOG'


class Actor:
  def __init__(self, name='', color=None, actor_id=1,
               x=0, y=0, alpha=0, beta=0,
               speed=0, period=0, steps=0,
               life_state=ADULT_STATE, baby_steps=0):
    self.name = name
    self.color = color if color is not None else ct.LIGHT_BLUE
    actor_id = int(actor_id)
    assert actor_id > 0, "actor_id should be greater than 0"
    self.id = int(actor_id)
    self.x = x
    self.y = y
    self.alpha = alpha
    self.beta = beta
    self.speed = speed
    self.period = period
    self.steps = steps
    self.life_state = life_state
    self.baby_steps = baby_steps

  def get_life_state_name(self):
    return LIFE_STATES[self.life_state]

  def get_data_string(self):
    return f'{self.id}, {self.name}, {self.get_life_state_name()}'

  def _change_direction(self, h=0, w=0):
    # alpha si beta sunt de fapt un R * sin, respectiv R * cos
    # deci o rotatie nu poate sa fie chiar solutia ideala, pt ca trebuie sa luam in calcul
    # unghiul curent.
    rad = random.uniform(0, 2 * math.pi)
    s, c = math.sin(rad), math.cos(rad)

    radius = math.sqrt(self.alpha * self.alpha + self.beta * self.beta)

    alpha = s * radius
    beta = c * radius

    if w != 0:
      beta = w * abs(beta)
    if h != 0:
      alpha = h * abs(alpha)

    self.alpha, self.beta = alpha, beta

  def change_direction(self):
    self._change_direction()
    self.steps = 0

  def _future_position(self, speed=None):
    if speed is None:
      speed = self.speed
    x = self.x + speed * self.alpha
    y = self.y + speed * self.beta

    return x, y

  def _step_forward(self, speed=None):
    self.x, self.y = self._future_position(speed)

  def _out_of_map(self, h, w, x=None, y=None):
    if x is None:
      x, y = self.x, self.y

    if x < 0 or y < 0:
      return True

    if h is not None and x >= h - EPSILON:
      return True
    if w is not None and y >= w - EPSILON:
      return True

    return False

  def _get_baby_steps_limit(self):
    return BABY_STEPS_LIMIT

  def make_adult(self):
    self.life_state = ADULT_STATE

    if self.speed == 0:
      self.speed = 1

    if self.alpha == 0 and self.beta == 0:
      self.alpha, self.beta = 0, 1

  def retire(self):
    self.life_state = RETIRED_STATE
    self.speed = 0
    self.alpha = 0
    self.beta = 0
    self.color = (100, 100, 100)

  def _check_baby_state(self):
    if self.life_state == BABY_STATE:
      self.baby_steps += 1
      if self.baby_steps > BABY_STEPS_LIMIT:
        self.make_adult()

      return True
    else:
      return False

  def check_same_direction(self):
    if self.period > 0 and self.steps % self.period < 1:
      self.change_direction()

  def still_in_map(self, x, y, h, w, speed=None):
    if not self._out_of_map(h, w, x, y):
      return x, y
    else:
      if speed is None:
        speed = self.speed
      tried = 0
      while self._out_of_map(h, w, x, y) and tried < BACK_TRIES_LIMIT:
        ch, cw = 0, 0
        if x < 0:
          ch = 1
        if h is not None and x >= h - EPSILON:
          ch = -1

        if y < 0:
          cw = 1
        if w is not None and y >= w - EPSILON:
          cw = -1

        self._change_direction(h=ch, w=cw)

        x, y = self._future_position(speed)
        tried += 1

      if tried < BACK_TRIES_LIMIT:
        self.steps = 0
        return x, y
      else:
        self.retire()
        return None, None

  def manage_collision(self, actor, env, x, y):
    self.change_direction()
    actor.change_direction()

  def forward(self, h=None, w=None, env=None):
    if self._check_baby_state():
      return

    if self.life_state == RETIRED_STATE:
      return

    self.check_same_direction()

    x, y = self._future_position()

    x, y = self.still_in_map(x, y, h, w)

    if x is not None:
      if env is not None:
        p_x, p_y = self.matrix_position(h, w, x, y)
        mat = env.get_current_map()

        current_value = mat[p_x, p_y]
        if current_value != 0:
          if current_value != self.id:
            if current_value > 0:
              # another actor
              actor = env.get_actor(current_value)
              self.manage_collision(actor, env, p_x, p_y)
            else:
              self.change_direction()
              # self.retire()
      # endif

      if self.life_state != RETIRED_STATE:
        self._step_forward()
        self.steps += 1

  def matrix_position(self, h=None, w=None, x=None, y=None):
    if x is None:
      x, y = self.x, self.y

    ans_x, ans_y = max(0, x), max(0, y)

    if w is not None:
      ans_y = min(ans_y, w - 1)
    if h is not None:
      ans_x = min(ans_x, h - 1)

    return math.floor(ans_x), math.floor(ans_y)

  def get_coords(self, h=None, w=None):
    x, y = self.matrix_position()
    xmin, xmax = x - SIZE, x + SIZE
    ymin, ymax = y - SIZE, y + SIZE
    xmin, ymin = max(xmin, 0), max(ymin, 0)

    if w is not None:
      ymax = min(ymax, w)
    if h is not None:
      xmax = min(xmax, h)

    return xmin, xmax, ymin, ymax

  def get_points(self, h=None, w=None):
    return [self.matrix_position(h, w)]

  def get_color(self):
    return self.color

  def get_id(self):
    return self.id

  def _head_collision(self, other):
    x1, y1 = self.matrix_position()
    x2, y2 = other.matrix_position()

    return x1 == x2 and y1 == y2

  def check_collision(self, other):
    return self._head_collision(other)




import numpy as np
import cv2
import seaborn as sns

from naeural_core import Logger
from decentra_vision.draw_utils import DrawUtils

if __name__ =='__main__':
  
  l = Logger('TST', base_folder='.', app_folder='_local_cache', TF_KERAS=False)
  
  painter = DrawUtils(log=l)
  
  cfg = l.load_data_json('config/config.txt')
  
  colors = cfg['CONFIG_PLUGINS']['HEATMAP_GRID_01']['HEAT_RANGE']
  
  img = np.zeros((1000,1000,3), dtype=np.uint8)
  
  for i, (start, stop, color) in enumerate(colors):
    img = painter.rectangle_tlbr(
      image=img, 
      top=i*100, 
      left=i*100, 
      bottom=(i+1)*100, 
      right=(i+1)*100,
      color=color,
      thickness=-1
      )
  cv2.imshow('Img', img)
  key = cv2.waitKey(0)
  cv2.close()





import argparse

from naeural_core import Logger
from naeural_core.local_libraries.model_server_v2.gateway import FlaskGateway

### Example for running a gateway
if __name__ == '__main__':
  parser = argparse.ArgumentParser()

  parser.add_argument(
    '-b', '--base_folder',
    type=str, default='libraries',
    help='Logger base folder'
  )

  parser.add_argument(
    '-a', '--app_folder',
    type=str, default='_logger_cache',
    help='Logger app folder'
  )

  parser.add_argument(
    '--host', type=str, default='0.0.0.0'
  )

  parser.add_argument(
    '--port', type=int, default=5002
  )

  args = parser.parse_args()
  base_folder = args.base_folder
  app_folder = args.app_folder
  host = args.host
  port = args.port

  try:
    ### Attention! config_file should contain the configuration for each endpoint; 'NR_WORKERS' and upstream configuration
    log = Logger(
      lib_name='GTW',
      config_file='libraries/model_server_v2/gateway/config_gateway.txt',
      base_folder=base_folder, app_folder=app_folder,
      TF_KERAS=False
    )

    gtw = FlaskGateway(
      log=log,
      server_names=['fake'],
      workers_location='libraries.model_server_v2.example_endpoints',
      workers_suffix='Worker',
      host=host,
      port=port,
      first_server_port=port+1,
      server_execution_path='/analyze'
    )
  except Exception as exc:
    print("FAILED to load config_gateway.txt: {}".format(exc))



"""
TODO: @Bleo use this and maybe "takeover"
"""

import naeural_core.constants as ct
import os

FN_PERSISTENT_DB = 'db_upload_mixin.pickle'

class _UploadMixin(object):
  def __init__(self):
    self._file_system_manager = None
    self.__um_initialized = False
    self.__uploaded_files = None

    super(_UploadMixin, self).__init__()
    return

  def _maybe_init_file_system_manager(self):
    if not self.__um_initialized:
      self._file_system_manager = self.global_shmem['file_system_manager']
      self.__um_initialized = True
      self.__uploaded_files = self.log.load_pickle_from_data(FN_PERSISTENT_DB, verbose=False) or set()
    #endif
    return

  def _save_db(self):
    self.log.save_pickle_to_data(self.__uploaded_files, FN_PERSISTENT_DB, verbose=False)
    return

  def upload_file(self, file_path, target_path, force_upload=False, save_db=False, verbose=0, **kwargs):
    self._maybe_init_file_system_manager()
    if self._file_system_manager is None:
      return None, None

    bool_new_file = file_path not in self.__uploaded_files
    # replace `\` with `/` on Windows, since this path is likely created with os.path.join
    target_path = target_path.replace(os.sep, '/')

    url = None
    if force_upload or bool_new_file:

      dct_upload_config = self.config_data.get(ct.IPFS.UPLOAD_CONFIG, {})
      kwargs = {**kwargs, **dct_upload_config}
      self.P("Uploading local file '{}' to remote '{}' {}...".format(
        file_path, target_path, "with overrides: {}".format(kwargs) if len(kwargs) >0 else '')
      )
      url = self._file_system_manager.upload(file_path, target_path, **kwargs)
    #endif

    if url is not None:
      if verbose >= 1:
        self.P("  Successfully uploaded.")
      self.__uploaded_files.add(file_path)
    #endif

    if save_db:
      self._save_db()

    return url, bool_new_file

  def upload_folder(self, folder_path, target_path, force_upload=False, **kwargs):
    # TODO: add include_subfolders and call this method from upload_output
    self._maybe_init_file_system_manager()
    if self._file_system_manager is None:
      return None

    urls = []
    _msg = "Uploading local folder '{}' to remote '{}' ...".format(folder_path, target_path)
    if force_upload:
      _msg += ' [FORCE]'
    self.P(_msg)

    files = os.listdir(folder_path)
    nr_old = 0
    for fn in files:
      url, bool_new_file = self.upload_file(
        file_path=os.path.join(folder_path, fn),
        target_path=os.path.join(target_path, fn),
        force_upload=force_upload,
        save_db=False,
        verbose=1,
        **kwargs
      )

      nr_old += not bool_new_file

      if url:
        urls.append(url)
    #endfor

    self.P("OK:{} old:{} total:{} files".format(len(urls), nr_old, len(files)))
    self._save_db()
    return urls

  def upload_logs(self, target_path, force_upload=False):
    if self._file_system_manager is None:
      return None

    urls = self.upload_folder(
      folder_path=self.log.get_target_folder('logs'),
      target_path=target_path,
      force_upload=force_upload
    )

    return urls

  def upload_output(self, target_path, force_upload=False):
    if self._file_system_manager is None:
      return None

    urls = []
    nr_chars_base_folder = len(self.log.get_base_folder())

    _msg = "Recursively uploading local cache output folder to remote '{}' ...".format(target_path)
    if force_upload:
      _msg += ' [FORCE]'
    self.P(_msg)

    nr_old = 0
    nr_total_files = 0
    for root, dirs, files in os.walk(self.log.get_target_folder('output')):
      nr_total_files += len(files)
      for file in files:
        url, bool_new_file = self.upload_file(
          file_path=os.path.join(root, file),
          target_path=os.path.join(
            target_path,
            '{}'.format(os.sep).join(root[nr_chars_base_folder+1:].split(os.sep)[1:]),
            file
          ),
          force_upload=force_upload,
          save_db=False,
          verbose=1
        )

        nr_old += not bool_new_file

        if url:
          urls.append(url)
      #endfor
    #endfor

    self.P("OK:{} old:{} total:{} files".format(len(urls), nr_old, nr_total_files))
    self._save_db()
    return urls


if __name__ == '__main__':

  from naeural_core import DecentrAIObject

  class Base(DecentrAIObject):
    def __init__(self, log, **kwargs):
      aaaa = 0
      super(Base, self).__init__(log=log, **kwargs)
      return

  class A(Base, _UploadMixin):
    def __init__(self, log, **kwargs):
      self._config_upload = {
        "TYPE" : "dropbox",
        "CONFIG_UPLOADER" : {
            "ACCESS_TOKEN": "4fohwiv544MAAAAAAAAAARvKOS34PxaMznuhdDZvTrT00zK1Sl9-W2TZgcUFbv9w"
        }
      }
      super(A, self).__init__(log=log, **kwargs)
      return


  from naeural_core import Logger

  log = Logger(lib_name='DBX', base_folder='.', app_folder='_local_cache', TF_KERAS=False)

  a = A(log=log)

  a.upload_output(target_path='_TEST_OUTPUT')




import tensorflow as tf
import os
import numpy as np

class AdvancedCallback(tf.keras.callbacks.Callback):
  """
      'Swiss Army Callback Knife'
  """

  def __init__(self,
               log,
               model_name=None,
               lr_monitor='val_loss',
               lr_mode='min',
               lr_factor=0.1,
               lr_patience=3,
               lr_min_delta=1e-4,
               lr_min_lr=1e-7,
               save_monitor='val_loss',
               save_mode='min',
               save_weights=False,
               save_last=2,
               metric_calc=None,
               calc_MAPE_TS=False,
               calc_MAD_TS=None,
               allow_neg_test=False,
               log_validate_each=1,
               nr_series_debug=None,
               y_trans=None,
               y_scale=None,
               DEBUG=False,
               **kwargs):
    """
    model_name : if None the model name will be inferred from tf.keras.models.Model.name
    lr_* : parameters for learning rate monitor

    save_monitor : what should be measured
    save_mode : how the checkpoint monitor should be measured
    save_last : the monitor will only keep best 'save_last' files

    metric_calc: function with signature func(model, data1, data2) -> dict
                 with "metric":value pairs

    y_scale, y_tras: transform y if needed

    log_validate_each :  will log chosen validation metrics every x epochs (default 1)

    calc_MAPE_TS: (default None) used in TS prediction calcs MAPE on last
                  time-step on validation data
    calc_MAD_TS: (default None) value x used in TS prediction to calculate
                  "period average deviation" for last x steps period
    *****

    `log` :  obiectul Logger din care se apeleaza - este automat dat de Logger - dont bother
    `model_name=None` : optional numele modelului urmarit - va lua numele Keras if None
    `lr_monitor='val_loss'` : ce monitorizam pentru scaderea ratei de invatare
    `lr_mode='min'` : cum monitorizam - min inseamna ca urmarim minimizarea lui lr_monitor
    `lr_factor=0.1` : cu ce scalam rata daca nu se respecta lr_mode pentru lr_patience pasi
    `lr_patience=3` : cati pasi avem 'rabdare' cu rata curenta
    `lr_min_delta=1e-4` : care este minimum de scadere/crestere a lr_monitor
    `lr_min_lr=1e-7`  : pana unde scadem de mult rata
    `save_monitor='val_loss'` : ce monitorizam ca sa salvam cele mai bune epoci
    `save_mode='min'` : cum monitorizam ...
    `save_last=2` : top epoci salvate
    `metric_calc=None` :  functie cu semnatura f(model,x_val,y_val)->dict(metric:val)
    `calc_MAPE_TS=False` : calcularea val_MAPE pe ultimul pas al series - doar pentru TS
    `calc_MAD_TS=None` : calculeaza val_MAD pe ultimii `calc_MAD_TS` pasi din TS
    `log_validate_each=1` : printeaza date validare la fiecare nr de epoci
    `y_trans=None` : variabila de translatare a output-ului de validare (daca a fost transf inainte)
    `y_scale=None` : variabile de scalare --//----
    `DEBUG=False` : True ca sa arate informatii extra la fiecare epoca
    `allow_neg_test=False` : default va face pozitive toate valorile pentru teste
    `save_weights=False` : true ca sa salveze doar weights

    """
    super().__init__(**kwargs)
    self.DEBUG = DEBUG
    self.K = tf.keras.backend

    if lr_factor >= 1.0:
      raise ValueError('ReduceLROnPlateau ' 'does not support a factor >= 1.0.')
    self.log = log
    self.__version__ = '1.1.7'
    self.model_name = model_name
    self.name = "ACb[{}]".format(self.model_name)
    self.lr_monitor = lr_monitor
    self.lr_factor = lr_factor
    self.lr_patience = lr_patience
    self.lr_min_delta = lr_min_delta
    self.lr_mode = lr_mode
    self.lr_min_lr = lr_min_lr
    self.log_validate_each = log_validate_each
    self.y_scale = y_scale
    self.y_trans = y_trans
    self.history = {}
    self.recorded_epochs = 0
    self.lr_monitor_op = None
    self.validation_funcs = {}
    self.save_monitor = save_monitor
    self.save_mode = save_mode
    self.save_last = save_last
    self.save_history = []
    self.allow_neg_test=allow_neg_test
    self.nr_series_debug = nr_series_debug
    self.__x_val = None
    self.__y_val = None
    self.save_weights = save_weights

    if metric_calc:
      func_name = metric_calc.__name__
      self.validation_funcs[func_name] = metric_calc

    if calc_MAPE_TS:
      func_name = "MAPE LAST STEP"
      self.validation_funcs[func_name] = self.calc_MAPE_TS

    if calc_MAD_TS:
      self.MAD_last_steps = calc_MAD_TS
      func_name = "MAD LAST {} STEPS".format(calc_MAD_TS)
      self.validation_funcs[func_name] = self.calc_MAD_TS

    self._reset()
    self.PrintSelf()
    return

  def PrintSelf(self):
    self.P(" v{}: {}".format(self.__version__, self.log.get_object_params(self)))

  def P(self, s, t=False, prefix=''):
    return self.log.P("{}{}: {}".format(
        prefix, self.name,s),show_time=t)


  def D(self, s, t=False, prefix=''):
    _r = -1
    if self.DEBUG:
      _r = self.log.P("{}[DEBUG] {}: {}".format(
                      prefix, self.name,s),show_time=t)
    return _r


  def _save_model(self, config_json=None):
    del_hist = max(1, self.save_last)-1
    for fn in self.save_history[:-del_hist]:
      if os.path.isfile(fn):
        try:
          os.remove(fn)
        except:
          pass
    label = "{}_ep{:03d}_{}_{:.3f}".format(self.model_name,
                                           self.recorded_epochs,
                                           self.save_monitor,
                                           self.save_best)

    self.last_save = self.log.save_keras_model(model=self.model,
                                               label=label,
                                               use_prefix=True,
                                               cfg=config_json,
                                               DEBUG=self.DEBUG,
                                               only_weights=self.save_weights)

    self.save_history.append(self.last_save)
    return


  def _reset(self):
    self.P("Resetting AdvancedCallback...")
    self.recorded_epochs = 0
    self.history = {}
    self.train_sessions = []
    self.last_save = None
    if self.lr_mode not in ['min', 'max']:
      raise ValueError("Unknown lr monitoring mode")
    if self.save_mode not in ['min', 'max']:
      raise ValueError("Unknown save monitoring mode")

    if self.lr_mode == 'min':
      self.lr_monitor_op = lambda a, b: np.less(a, b - self.lr_min_delta)
      self.lr_best = np.Inf
    else:
      self.lr_monitor_op = lambda a, b: np.greater(a, b + self.lr_min_delta)
      self.lr_best = -np.Inf

    if self.save_mode == 'min':
      self.save_monitor_op = lambda a, b: np.less(a, b - self.lr_min_delta)
      self.save_best = np.Inf
    else:
      self.save_monitor_op = lambda a, b: np.greater(a, b + self.lr_min_delta)
      self.save_best = -np.Inf

    self.cooldown_counter = 0
    self.wait = 0
    self.n_vals = len(self.validation_funcs)
    return


  def _lr_monitor(self, epoch, logs):
    if (self.lr_monitor is None) or (self.lr_monitor == ''):
      # we dont need lr monitoring
      return
    current = logs.get(self.lr_monitor)
    if current is None:
      raise ValueError("Learning rate monitoring metric {} missing!".format(
          self.lr_monitor))
    else:
      if self.lr_monitor_op(current, self.lr_best):
        _outcome = 'decreased'
        self.lr_best = current
        self.wait = 0
      else:
        _outcome = 'increased'
        self.wait += 1
        if self.wait >= self.lr_patience:
          old_lr = float(self.K.get_value(self.model.optimizer.lr))
          if old_lr > self.lr_min_lr:
            new_lr = old_lr * self.lr_factor
            new_lr = max(new_lr, self.lr_min_lr)
            self.K.set_value(self.model.optimizer.lr, new_lr)
            self.P("Epoch {}({}): {} {} from {:.4f} to {:.4f} ".format(
                              epoch + 1, self.recorded_epochs,
                              self.lr_monitor, _outcome,
                              self.lr_best, current,), prefix='\n')
            self.P('             => reducing lr from {:.7f} to {:.7f}'.format(
                          old_lr, new_lr))
            self.wait = 0
            self.lr_best = current # reset the best
    return


  def _checkpoint_monitor(self, epoch, logs):
    if (self.save_monitor is None) or (self.save_monitor == ''):
      # no checkpoint
      return
    current = logs.get(self.save_monitor)
    if current is None:
      raise ValueError("Checkpoint save monitoring metric {} missing!".format(
          self.save_monitor))
    else:
      if self.save_best > current:
        _outcome = 'decreased'
      else:
        _outcome = 'increased (NO SAVING)'
      self.D("Epoch {}({}): {} {} from {:.4f} to {:.4f} ".format(
                         epoch + 1, self.recorded_epochs,
                         self.save_monitor, _outcome,
                         self.save_best, current,), prefix='\n')
      if self.save_monitor_op(current, self.save_best):
        self.save_best = current
        self._save_model()
    return


  def on_train_begin(self, logs=None):
    if self.model_name is None:
      self.model_name = self.model.name
    self.nr_inputs = len(self.model.inputs)
    self.nr_outputs = len(self.model.outputs)
    self.output_names = [x.name.split('/')[0] for x in self.model.outputs]
    self.name = "ACb[{}]".format(self.model_name)
    self.train_sessions.append(
          {self.name : 0}
        )
    c_lr = self.K.get_value(self.model.optimizer.lr)
    prefix = ""
    if self.recorded_epochs > 0:
      prefix = "(Re)"
    self.P("{}Starting train with lr={:.7f}. So far trained {} epochs".format(
        prefix, c_lr, self.recorded_epochs))
    return


  def on_epoch_end(self, epoch, logs=None):
    if (self.n_vals > 0)  and (self.validation_data) and (self.__x_val == None):
      # setup validation only if required by validation functions
      # so check if we have functions, input data and if we did not setup already
      if ((type(self.validation_data) == tuple) and (len(self.validation_data) == 2)
           and self.nr_inputs > 1):
        # now we have to re-arrange the validation data
        val_data = [x for x in self.validation_data[0]]
        if self.validation_data[1] is not None:
          val_data += [x for x in self.validation_data[1]]
        self.validation_data = val_data
      self.__x_val = self.validation_data[:self.nr_inputs]
      self.__y_val = self.validation_data[self.nr_inputs:(self.nr_inputs+self.nr_outputs)]
      if len(self.__x_val) == 1:
        self.__x_val = self.__x_val[0]
      if len(self.__y_val) == 1:
        self.__y_val = self.__y_val[0]
      self.P("")
      self.P("Validation at ep {} on data len={}".format(self.recorded_epochs,
                                                         len(self.validation_data)))
      yv = self.__y_val
      xv = self.__x_val
      if type(yv) in [np.ndarray, list, tuple]:
        yvs = yv.shape if type(yv) == np.ndarray else [y.shape for y in yv
                                                      if type(y) == np.ndarray]
      if type(xv) in [np.ndarray, list, tuple]:
        xvs = xv.shape if type(xv) == np.ndarray else [x.shape for x in xv
                                                      if type(x) == np.ndarray]
      self.P(" y_val={}".format(yvs))
      self.P(" x_val={}".format(xvs))

    self.train_sessions[-1][self.name] += 1
    self.recorded_epochs += 1
    logs = logs or {}
    # this will update global log
    logs['lr'] = self.K.get_value(self.model.optimizer.lr)

    if (self.recorded_epochs % self.log_validate_each) == 0:
      self.P("")
      self.P("Validation tests on model '{}' at epoch {}".format(
          self.model_name, self.recorded_epochs))
    for func_name in self.validation_funcs:
      func_val = self.validation_funcs[func_name]
      assert self.__y_val is not None, "Missing validation data for validation func '{}'!"\
      " This is probably due to the version of keras/tensorflow that removed the "\
      "validation_data variables setup. Please assign validation_data manually".format(
                          func_name)
      dict_res = func_val(self.model, self.__x_val, self.__y_val)
      for res_key in dict_res:
        # this will update global log
        logs[res_key] = dict_res[res_key]

    for log_key in logs:
      key_hist = self.history.get(log_key) or []
      key_hist.append(logs[log_key])
      self.history[log_key] = key_hist

    self._lr_monitor(epoch=epoch, logs=logs)
    self._checkpoint_monitor(epoch=epoch, logs=logs)
    return


  def calc_MAPE_TS(self, model, x_val, y_val, stand_alone=False):
    y_pred = model.predict(x_val)
    if type(y_val) == list:
      y_pred = np.concatenate(y_pred, axis=-1)
      y_val = np.concatenate(y_val, axis=-1)
    if self.y_scale:
      self.P("  Scaling y data by {}".format(self.y_scale))
      y_pred = y_pred * self.y_scale
      y_val = y_val * self.y_scale
    if self.y_trans:
      self.P("  Translating y data by {}".format(self.y_trans))
      y_pred = y_pred + self.y_trans
      y_val = y_val + self.y_trans

    if not self.allow_neg_test:
      y_val = np.clip(y_val,self.K.epsilon(), None)
      y_pred = np.clip(y_pred, self.K.epsilon(), None)

    y_pred_last = y_pred[:,-1,:]
    y_val_last = y_val[:,-1,:]

    residual = y_pred_last - y_val_last

    if hasattr(model,"loss"):
      loss = model.loss
    else:
      loss = "MSE"

    loss_type = 'MSE'
    if isinstance(loss, list):
      loss = str(loss[0])
    if 'MAE' in loss.upper():
      loss_type = 'MAE'
      MAPE_loss = np.mean(np.abs(residual))
    elif 'LOG' in loss.upper():
      loss_type = 'LOGCOSH'
      MAPE_loss = np.mean(np.log(np.cosh(residual)))
    else:
      # MSE
      MAPE_loss = np.mean(residual ** 2)

    self.P("  Using {} for val_MAPE_loss calculation".format(loss_type))
    mape_series = np.abs(residual) / (
                  np.clip(np.abs(y_val_last), self.K.epsilon(), None))
    mape_series = np.clip(mape_series, 0, 2)

    if self.nr_series_debug:
      nr_series = self.nr_series_debug
    else:
      nr_series = min(10, mape_series.shape[0])

    mape = np.mean(mape_series)
    res_dict = {"val_MAPE" : mape, "val_MAPE_loss" : MAPE_loss}
    self.P("  Using {} for val_MAPE_loss={:.4f}".format(loss_type, MAPE_loss))
    if (self.recorded_epochs % self.log_validate_each) == 0:
      self.P(" Validation MAPE (last step) on {} outputs {}:".format(
          y_val.shape[-1], self.output_names))
      self.P("  y_val({} series): {}".format(nr_series,
                                            y_val[:nr_series,-1,:].ravel().round(2)))
      self.P("  y_hat({} series): {}".format(nr_series,
                                            y_pred[:nr_series,-1,:].ravel().round(2)))
      self.P("  MAPEs({} series): {}".format(nr_series,
                                            mape_series[:nr_series].ravel().round(2)))
      self.P("  Mean Average Percentual ERROR: {:.1f}%".format(mape*100))

    if not stand_alone:
      _return = res_dict
    else:
      _return = res_dict, mape_series

    return _return


  def calc_MAD_TS(self, model, x_val, y_val):
    assert self.MAD_last_steps > 0
    lstp = self.MAD_last_steps
    y_pred = model.predict(x_val)
    if type(y_val) == list:
      y_pred = np.concatenate(y_pred, axis=-1)
      y_val = np.concatenate(y_val, axis=-1)
    assert len(y_val.shape) == 3, "The time-series must be in (B,T,F) format!"

    sum_axis = 1
    if (lstp == 1) and (y_val.shape[-1] > 1):
      sum_axis = 2

    if self.y_scale:
      self.P("  Scaling y data by {}".format(self.y_scale))
      y_pred = y_pred * self.y_scale
      y_val = y_val * self.y_scale
    if self.y_trans:
      self.P("  Translating y data by {}".format(self.y_trans))
      y_pred = y_pred + self.y_trans
      y_val = y_val + self.y_trans

    if not self.allow_neg_test:
      y_val = np.clip(y_val,self.K.epsilon(), None)
      y_pred = np.clip(y_pred, self.K.epsilon(), None)

    yp = y_pred[:,-lstp:,:]
    yt = y_val[:,-lstp:,:]
    yp_s = yp.sum(axis=sum_axis)
    yt_s = yt.sum(axis=sum_axis)
    yp_s = np.clip(yp_s, self.K.epsilon(), None)
    np_mad = np.abs(yp_s-yt_s) / yp_s
    np_mad = np.clip(np_mad, 0, 2)
    mad = np_mad.mean()

    if self.nr_series_debug:
      nr_series = self.nr_series_debug
    else:
      nr_series = min(2, np_mad.shape[0])

    res_dict = {"val_MAD" : mad}
    if (self.recorded_epochs % self.log_validate_each) == 0:
      yvl = "".join(['{}'.format(x.ravel().round(2)) for x in y_val[:nr_series,-lstp:,:]])
      yhl = "".join(['{}'.format(x.ravel().round(2)) for x in y_pred[:nr_series,-lstp:,:].astype('float64')])
      self.P(" Validation MAD last {} steps on {} outputs {}:".format(
           self.MAD_last_steps, y_val.shape[-1], self.output_names,))
      self.P("  y_val({} series): {}".format(nr_series, yvl))
      self.P("  y_hat({} series): {}".format(nr_series, yhl))
      self.P("  MADs({} series):  {}".format(nr_series,
                                            np_mad[:nr_series].ravel().round(2)))
      self.P("  Mean Period({}) Average Deviation ERROR: {:.1f}%".format(lstp, mad*100))
    return res_dict


import numpy as np
import torch as th

from naeural_core.local_libraries.nn.th.training.callbacks.classification import ClassificationTrainingCallbacks

class SiameseTrainingCallbacks(ClassificationTrainingCallbacks):
  def __init__(self, distance_multiplier=None, **kwargs):
    # Parameter used as a scalar multiplier for distance between classes; used in contrastive loss and triplet loss
    self.distance_multiplier = distance_multiplier
    super(SiameseTrainingCallbacks, self).__init__(**kwargs)


  def contrastive_train_on_batch(self, model, optimizer, losses, batch, y_index=1, batch_index=False):
    """
    Train on batch method used for training models with contrastive loss.

    Each batch must have the format: (anchor, test_image), distance

    It computes the embeddings for the anchors, then the test images and applies the loss
    """
    lst_th_x, lst_th_y = self._get_batch_lst_x_lst_y(batch=batch, y_index=y_index, batch_index=batch_index)

    (a, im) = lst_th_x
    a_enc, im_enc = model(a), model(im)

    losses_vals = [losses['loss'](a_enc, im_enc, lst_th_y)]
    th_loss = th.stack(losses_vals, dim=0).sum(dim=0)

    optimizer.zero_grad()
    th_loss.mean().backward()
    optimizer.step()
    err = th_loss.detach().cpu().numpy()
    return err

  def triplet_loss_train_on_batch(self, model, optimizer, losses, batch, y_index=1, batch_index=False):
    """
    Train on batch method used for training models with contrastive loss.

    Each batch must have the format: (anchor, close_image, distant_image), distance

    It computes the embeddings for the anchors, for the close images and for the distant images and finally applies the loss
    """
    lst_th_x, lst_th_y = self._get_batch_lst_x_lst_y(batch=batch, y_index=y_index, batch_index=batch_index)

    (a, p, n) = lst_th_x
    a_enc, p_enc, n_enc = model(a), model(p), model(n)

    losses_vals = [losses['loss'](a_enc, p_enc, n_enc, lst_th_y)]
    th_loss = th.stack(losses_vals, dim=0).sum(dim=0)

    optimizer.zero_grad()
    th_loss.mean().backward()
    optimizer.step()
    err = th_loss.detach().cpu().numpy()
    return err

  def _contrastive_loss_forward_and_collect(self, data_generator: th.utils.data.DataLoader, dataset_info : dict, **kwargs):
    with th.no_grad():
      breakpoints = self.get_breakpoints(model_class=kwargs.get('model_class', 'ContrastiveLoss'))
      steps_per_epoch = len(data_generator)
      iterator = iter(data_generator)
      lst_a_enc, lst_im_enc, lst_y, lst_indexes = [], [], [], []
      for i in range(steps_per_epoch):
        x_batch, y_batch, idx_batch = next(iterator)
        x_batch, y_batch = self._transform_at_eval_time(x_batch, y_batch)
        a, im = x_batch
        a_enc, im_enc = self._model(a), self._model(im)
        lst_a_enc.append(a_enc.cpu().numpy())
        lst_im_enc.append(im_enc.cpu().numpy())
        lst_y.append(y_batch.cpu().numpy())

        lst_indexes.append(idx_batch)

        if self.log.is_main_thread:
          print(
            '\r  Forward progress: {}/{} ({:.2f}%)'.format(i + 1, steps_per_epoch, 100 * (i + 1) / steps_per_epoch),
            flush=True, end=''
          )
      #endfor
      a_encs = np.vstack(lst_a_enc)
      im_encs = np.vstack(lst_im_enc)
      dist = im_encs - a_encs
      l1_dist = np.linalg.norm(dist, axis=1, ord=1)

      y_hat = np.argmin(np.abs(l1_dist.reshape((-1, 1)) - breakpoints), axis=-1)
      y = np.hstack(lst_y)
      idx = self._get_idx(lst_indexes)
    #endwith
    return y, y_hat, idx

  def _triplet_loss_forward_and_collect(self, data_generator: th.utils.data.DataLoader, dataset_info : dict, **kwargs):
    with th.no_grad():
      breakpoints = self.get_breakpoints(model_class=kwargs.get('model_class', 'TripletLoss'))
      steps_per_epoch = len(data_generator)
      iterator = iter(data_generator)
      lst_a_enc, lst_p_enc, lst_n_enc, lst_y, lst_indexes = [], [], [], [], []
      for i in range(steps_per_epoch):
        x_batch, y_batch, idx_batch = next(iterator)
        x_batch, y_batch = self._transform_at_eval_time(x_batch, y_batch)
        a, p, n = x_batch
        a_enc, p_enc, n_enc = self._model(a), self._model(p), self._model(n)

        lst_a_enc.append(a_enc.cpu().numpy())
        lst_p_enc.append(p_enc.cpu().numpy())
        lst_n_enc.append(n_enc.cpu().numpy())
        lst_y.append(y_batch.cpu().numpy())

        lst_indexes.append(idx_batch) ## THIS DOESN"T WORK

        if self.log.is_main_thread:
          print(
            '\r  Forward progress: {}/{} ({:.2f}%)'.format(i + 1, steps_per_epoch, 100 * (i + 1) / steps_per_epoch),
            flush=True, end=''
          )
      #endfor
      a_encs = np.vstack(lst_a_enc * 2)
      im_encs = np.vstack(lst_p_enc + lst_n_enc)
      dist = im_encs - a_encs
      l1_dist = np.linalg.norm(dist, axis=1, ord=1)

      y_hat = np.argmin(np.abs(l1_dist.reshape((-1, 1)) - breakpoints), axis=-1)
      y = np.hstack([np.array(0) for _ in range(im_encs.shape[0] // 2)] + lst_y)

      idx = self._get_idx(lst_indexes * 2)
    #endwith
    return y, y_hat, idx


  def get_breakpoints(self, model_class, number_of_classes=5):
    """
    Heuristic used for computing the median distances between classes. Used for inference of models trained by using
      triplet loss or contrastive loss.

    Works by running inference on the train set and then finding the median l1 distance between each class and the anchors

    At inference time y_hat is equals to the class that has its breakpoint (median) closest to the inferred distance to
      the anchor

    """
    th_dl = self._owner.th_dl
    assert model_class in ['ContrastiveLoss', 'TripletLoss'], ValueError("Invalid model class ''".format(model_class))

    self.P("Computing breakpoints ...")
    if model_class == 'ContrastiveLoss':
      lst_a, lst_im, lst_y = [], [], []
    elif model_class == 'TripletLoss':
      lst_a, lst_p, lst_n, lst_y = [], [], [], []
    #endif model_class

    iterator = iter(th_dl)
    steps_per_epoch = len(th_dl)
    for i in range(steps_per_epoch):
      x_batch, y_batch, idx_batch = next(iterator)
      x_batch, y_batch = self._transform_at_eval_time(x_batch, y_batch)

      if model_class == 'ContrastiveLoss':
        a, im = x_batch
        a_enc, im_enc = self._model(a), self._model(im)
        lst_a.append(a_enc.cpu().numpy())
        lst_im.append(im_enc.cpu().numpy())
        lst_y.append(y_batch.cpu().numpy())
      elif model_class == 'TripletLoss':
        a, p, n = x_batch
        a_enc, p_enc, n_enc = self._model(a), self._model(p), self._model(n)

        lst_a.append(a_enc.cpu().numpy())
        lst_p.append(p_enc.cpu().numpy())
        lst_n.append(n_enc.cpu().numpy())
        lst_y.append(y_batch.cpu().numpy())
      else:
        raise ValueError("Invalid model class '{}'".format(self._model.model_class))
      #end if model_class

      if self.log.is_main_thread:
        print(
          '\r  Forward progress: {}/{} ({:.2f}%)'.format(i + 1, steps_per_epoch, 100 * (i + 1) / steps_per_epoch),
          flush=True, end=''
        )
      #endif
    #endfor
    if model_class == 'ContrastiveLoss':
      a_encs = np.vstack(lst_a)
      im_encs = np.vstack(lst_im)
      ys = np.hstack(lst_y)
    elif model_class == 'TripletLoss':
      a_encs = np.vstack(lst_a * 2)
      im_encs = np.vstack(lst_p + lst_n)
      ys = np.hstack([np.array(0) for _ in range(im_encs.shape[0]//2)] + lst_y)

    dist_hat = im_encs - a_encs

    median_distances = np.array([
      np.median(np.linalg.norm(dist_hat, axis=1, ord=1)[ys == i * self.distance_multiplier])
      for i in range(number_of_classes)
    ])

    self.P("Breakpoints computed")
    self.P("TRAIN SET: Median l1 distances:\n {}".format(median_distances))

    return median_distances


import gc
import traceback
import platform

class _GPUMixin(object):
  """
  Mixin for GPU functionalities that are attached to `libraries.logger.Logger`.

  This mixin cannot be instantiated because it is built just to provide some additional
  functionalities for `libraries.logger.Logger`

  In this mixin we can use any attribute/method of the Logger.

  * Obs: This mixin uses also attributes/methods of `_MachineInfoMixin`:
    - self.get_machine_memory
    - self.get_avail_memory
  """

  def __init__(self):
    super(_GPUMixin, self).__init__()

    try:
      from ratio1.logging.logger_mixins.machine_mixin import _MachineMixin
    except ModuleNotFoundError:
      raise ModuleNotFoundError("Cannot use _GPUMixin without having _MachineMixin")

    self._done_first_smi_error = False
    return

  @staticmethod
  def clear_gpu_memory():
    try:
      import tensorflow as tf
      import gc
      gc.collect()
      tf.keras.backend.clear_session()
    except:
      pass
    try:
      import torch as th
      import gc
      gc.collect()
      th.cuda.empty_cache()
      th.cuda.clear_memory_allocated()
    except:
      pass

  @staticmethod
  def get_gpu_memory_map():
    import subprocess
    """Get the current gpu usage.

    Returns
    -------
    usage: dict
        Keys are device ids as integers.
        Values are memory usage as integers in MB.
    """
    result = subprocess.check_output(
      [
        'nvidia-smi', '--query-gpu=memory.used',
        '--format=csv,nounits,noheader'
      ])
    result = result.decode('utf-8')
    # Convert lines into a dictionary
    gpu_memory = [int(x) for x in result.strip().split('\n')]
    gpu_memory_map = dict(zip(range(len(gpu_memory)), gpu_memory))
    return gpu_memory_map


  def gpu_info(self, show=False, mb=False, current_pid=False):
    """
    Collects GPU info. Must have torch installed & non-mandatory nvidia-smi

    Parameters
    ----------
    show : bool, optional
      show data as gathered. The default is False.
    mb : bool, optional
      collect memory in MB otherwise in GB. The default is False.
    current_pid: bool, optional
      return data only for GPUs used by current process or all if current process does
    not use GPU
    

    Returns
    -------
    lst_inf : list of dicts
      all GPUs info from CUDA:0 to CUDA:n.

    """

    def _main_func():
      if vars(self).get('__no_gpu_avail', False):
        return []

      try:
        # first get name
        import torch as th
        import os
      except:
        self.P("ERROR: `gpu_info` call failed - PyTorch probably is not installed:\n{}".format(
          traceback.format_exc())
        )
        self.__no_gpu_avail = True
        return None
      

      nvsmires, device_props, dct_proc_info = None, None, None
      try:
        from pynvml_utils import nvidia_smi
        import pynvml
        nvsmi = nvidia_smi.getInstance()
        nvsmires = nvsmi.DeviceQuery('memory.free, memory.total, memory.used, utilization.gpu, temperature.gpu, fan.speed')
        pynvml_avail = True
      except:
        pynvml_avail = False
      
      if not pynvml_avail or len(nvsmires) == 0:
        self.__no_gpu_avail = True

      lst_inf = []
      # now we iterate all devices
      n_gpus = th.cuda.device_count()
      if n_gpus > 0:
        gc.collect()
        th.cuda.empty_cache()
      current_pid_has_usage = False
      current_pid_gpus = []

      try:
        for device_id in range(n_gpus):
          dct_device = {}
          device_props = th.cuda.get_device_properties(device_id)
          dct_device['NAME'] = device_props.name
          dct_device['TOTAL_MEM'] = round(
            device_props.total_memory / 1024 ** (2 if mb else 3),
            2
          )
          mem_total = None
          mem_allocated = None
          gpu_used = None
          gpu_temp = None
          gpu_temp_max = None
          fan_speed, fan_speed_unit = -1, "N/A"
          if pynvml_avail and nvsmires is not None and 'gpu' in nvsmires:
            dct_gpu = nvsmires['gpu'][device_id]
            mem_total = round(
              dct_gpu['fb_memory_usage']['total'] / (1 if mb else 1024),
              2
            )  # already from th
            mem_allocated = round(
              dct_gpu['fb_memory_usage']['used'] / (1 if mb else 1024),
              2
            )
            gpu_used = dct_gpu['utilization']['gpu_util']
            if isinstance(gpu_used, str):
              gpu_used = -1
            gpu_temp = dct_gpu['temperature']['gpu_temp']
            gpu_temp_max = dct_gpu['temperature']['gpu_temp_max_threshold']
            
            fan_speed = dct_gpu.get('fan_speed', -1)
            fan_speed = -1 if fan_speed == 'N/A' else fan_speed
            fan_speed_unit = dct_gpu.get('fan_speed_unit', "N/A")

            handle = pynvml.nvmlDeviceGetHandleByIndex(device_id)
            processes = []
            for proc in pynvml.nvmlDeviceGetComputeRunningProcesses(handle):
              dct_proc_info = {k.upper(): v for k,v in proc.__dict__.items()}
              used_mem = dct_proc_info.pop('USEDGPUMEMORY', None)
              dct_proc_info['ALLOCATED_MEM'] = round(
                used_mem / 1024 ** (2 if mb else 3) if used_mem is not None else 0.0,
                2
              )
              processes.append(dct_proc_info)
              if dct_proc_info['PID'] == os.getpid():
                current_pid_has_usage = True
                current_pid_gpus.append(device_id)
            #endfor
            dct_device['PROCESSES'] = processes
            dct_device['USED_BY_PROCESS'] = device_id in current_pid_gpus            
          else:
            str_os = platform.platform()
            ## check if platform is Tegra and record
            if 'tegra' in str_os.lower():
              # we just record the overall fre memory
              mem_total = self.get_machine_memory(gb=not mb)
              mem_allocated = mem_total  - self.get_avail_memory(gb=not mb)
              gpu_used = 1
              gpu_temp = 1
              gpu_temp_max = 100
              if not self._done_first_smi_error and nvsmires is not None:
                self.P("Running `gpu_info` on Tegra platform: {}".format(nvsmires), color='r')
                self._done_first_smi_error = True
            elif not self._done_first_smi_error:
              str_log = "ERROR: Please make sure you have both pytorch and pynvml in order to monitor the GPU"
              str_log += "\nError info: pynvml_avail={}, nvsmires={}".format(pynvml_avail, nvsmires)
              self.P(str_log)
              self._done_first_smi_error = True
          #endif
          dct_device['ALLOCATED_MEM'] = mem_allocated
          dct_device['FREE_MEM'] = -1
          if all(x is not None for x in [mem_total, mem_allocated]):
            dct_device['FREE_MEM'] = round(mem_total - mem_allocated, 2)
          dct_device['MEM_UNIT'] = 'MB' if mb else 'GB'
          dct_device['GPU_USED'] = gpu_used
          dct_device['GPU_TEMP'] = gpu_temp
          dct_device['GPU_TEMP_MAX'] = gpu_temp_max
          dct_device['GPU_FAN_SPEED'] = fan_speed 
          dct_device['GPU_FAN_SPEED_UNIT'] = fan_speed_unit

          lst_inf.append(dct_device)
        #end for all devices
      except Exception as e:
        self.P("gpu_info exception for device_id {}: {}\n devicepros: {}\n nvsmires: {}\n dct_proc_info: {}".format(
          device_id, e, 
          device_props, nvsmires, dct_proc_info), color='r'
        )

      if show:
        self.P("GPU information for {} device(s):".format(len(lst_inf)), color='y')
        for dct_gpu in lst_inf:
          for k, v in dct_gpu.items():
            self.P("  {:<14} {}".format(k + ':', v), color='y')

      if current_pid and current_pid_has_usage:
        return [lst_inf[x] for x in current_pid_gpus]
      else:
        return lst_inf
    #enddef

    # if multiple threads call at the same time the method `log.gpu_info`, then `pynvml_avail` will be False
    #   (most probably due to the queries that are performed using nvidia_smi)
    with self.managed_lock_resource('gpu_info'):
      res = _main_func()
    # endwith lock
    return res
  
  
  def get_gpu_info(self, device_id=0, mb=False):
    """
    Returns GPU info for a specific device
    """
    res = {}
    if not vars(self).get('__no_gpu_avail', False):
      gpu_info = self.gpu_info()
      if gpu_info is not None and len(gpu_info) > 0 and device_id < len(gpu_info):
        res = gpu_info[device_id]
    return res
  
  
  def get_gpu_total_mem(self, device_id=0, mb=False):
    """
    Returns total memory of a specific GPU device
    """
    res = 0
    if not vars(self).get('__no_gpu_avail', False):
      device_info = self.get_gpu_info(device_id=device_id, mb=mb)
      if device_info is not None and len(device_info) > 0:
        res = device_info['TOTAL_MEM']
    return res
  
  
  def get_gpu_free_mem(self, device_id=0, mb=False):
    """
    Returns free memory of a specific GPU device
    """
    res = 0
    if not vars(self).get('__no_gpu_avail', False):
      device_info = self.get_gpu_info(device_id=device_id, mb=mb)
      if device_info is not None and len(device_info) > 0:
        res = device_info['FREE_MEM']
    return res
  
  
  def get_gpu_name(self, device_id=0):
    """
    Returns name of a specific GPU device
    """
    res = 'N/A'
    if not vars(self).get('__no_gpu_avail', False):
      device_info = self.get_gpu_info(device_id=device_id)
      if device_info is not None and len(device_info) > 0:
        res = device_info['NAME']
    return res
  
  
  def get_gpu_fan_speed(self, device_id=0):
    """
    Returns fan speed of a specific GPU device
    """
    res = 'N/A'
    if not vars(self).get('__no_gpu_avail', False):
      device_info = self.get_gpu_info(device_id=device_id)
      if device_info is not None and len(device_info) > 0:
        res = device_info.get('GPU_FAN_SPEED')
    return res
    


if __name__ == '__main__':
  print(_GPUMixin.get_gpu_memory_map())



from naeural_core.business.base import BasePluginExecutor as BasePlugin

__VER__ = '0.1.0.0'

_CONFIG = {
  **BasePlugin.CONFIG,
  'VALIDATION_RULES': {
    **BasePlugin.CONFIG['VALIDATION_RULES'],
  },

  'ALLOW_EMPTY_INPUTS' : True,

  'DATASET_OBJECT_NAME' : None,
  'DATASET_LOCAL_PATH'  : None,

  'PLUGIN_LOOP_RESOLUTION' : 1/5,
}

class MinioDownloadDatasetPlugin(BasePlugin):
  CONFIG = _CONFIG
  def __init__(self, **kwargs):
    self._downloaded = False
    super(MinioDownloadDatasetPlugin, self).__init__(**kwargs)
    return

  def startup(self):
    super().startup()
    assert self.cfg_dataset_object_name is not None
    assert self.cfg_dataset_local_path is not None
    return

  @property
  def cfg_dataset_object_name(self):
    return self._instance_config['DATASET_OBJECT_NAME']

  @property
  def cfg_dataset_local_path(self):
    return self._instance_config['DATASET_LOCAL_PATH']

  def _process(self):
    if not self._downloaded:
      self.global_shmem['file_system_manager'].download(
        uri=self.cfg_dataset_object_name,
        local_file_path=self.cfg_dataset_local_path
      )
      self._downloaded = True
      self.cmdapi_stop_current_stream()
    #endif

    return



from naeural_core.remote_file_system.base import BaseFileSystem
from naeural_core import constants as ct

_CONFIG = {
  **BaseFileSystem.CONFIG,
  'VALIDATION_RULES': {
    **BaseFileSystem.CONFIG['VALIDATION_RULES'],
  },
}

class MinioFileSystem(BaseFileSystem):
  """
  configuration example:
    ...
    "CONFIG_UPLOADER" : 
      {
        "ENDPOINT"    : "MINIO_ENDPOINT",
        "ACCESS_KEY"  : "MINIO_ACCESS_KEY",
        "SECRET_KEY"  : "MINIO_SECRET_KEY",
        "BUCKET_NAME" : "MINIO_BUCKET_NAME",
      }
    ...
    
  """

  CONFIG = _CONFIG
  def __init__(self, **kwargs):
    super(MinioFileSystem, self).__init__(**kwargs)
    return

  def upload(self, file_path, target_path, **kwargs):
    dct_minio = {k.lower(): v for k,v in self.config_data.items()}
    
    access_key = kwargs.get('access_key') or dct_minio['access_key']
    secret_key = kwargs.get('secret_key') or dct_minio['secret_key']
    bucket_name = kwargs.get('bucket_name') or dct_minio['bucket_name']
    endpoint = kwargs.get('endpoint') or dct_minio['endpoint']
    secure = kwargs.get('secure') or dct_minio.get('secure', False)
    days_retention = kwargs.get('days_retention') or dct_minio.get('days_retention')
    
    url = self.log.minio_upload(
      file_path=file_path, 
      object_name=target_path,
      endpoint=endpoint, 
      access_key=access_key, 
      secret_key=secret_key, 
      bucket_name=bucket_name, 
      days_retention=days_retention,
      secure=secure,
    )
    return url

  def download(self, uri, local_file_path, **kwargs):
    if uri.startswith('http'):
      return self._http_download(url=uri, local_file_path=local_file_path)

    dct_minio = {k.lower(): v for k,v in self.config_data.items()}

    access_key = kwargs.get('access_key') or dct_minio['access_key']
    secret_key = kwargs.get('secret_key') or dct_minio['secret_key']
    bucket_name = kwargs.get('bucket_name') or dct_minio['bucket_name']
    endpoint = kwargs.get('endpoint') or dct_minio['endpoint']
    secure = kwargs.get('secure') or dct_minio.get('secure', False)

    res = self.log.minio_download(
      local_file_path=local_file_path,
      endpoint=endpoint,
      access_key=access_key,
      secret_key=secret_key,
      bucket_name=bucket_name,
      object_name=uri,
      secure=secure,
    )

    return res

  def list_objects(self, only_names=True, **kwargs):
    dct_minio = {k.lower(): v for k,v in self.config_data.items()}
    
    access_key = kwargs.get('access_key') or dct_minio['access_key']
    secret_key = kwargs.get('secret_key') or dct_minio['secret_key']
    bucket_name = kwargs.get('bucket_name') or dct_minio['bucket_name']
    endpoint = kwargs.get('endpoint') or dct_minio['endpoint']
    secure = kwargs.get('secure') or dct_minio.get('secure', False)

    from minio import Minio
    try:
      client = Minio(
        endpoint=endpoint,
        access_key=access_key,
        secret_key=secret_key,
        secure=secure,
      )
      lst_objects = list(client.list_objects(bucket_name=bucket_name, recursive=True))
      if only_names:
        lst_objects = list(map(lambda x: x.object_name, lst_objects))
    except:
      lst_objects = []

    return lst_objects
  
if __name__ == '__main__':
  from naeural_core import Logger
  cfg = {
        "ENDPOINT"    : "MINIO_ENDPOINT",
        "ACCESS_KEY"  : "MINIO_ACCESS_KEY",
        "SECRET_KEY"  : "MINIO_SECRET_KEY",
        "BUCKET_NAME" : "MINIO_BUCKET_NAME",
  }
  l = Logger('T', base_folder='.', app_folder='_local_cache')
  ttt = MinioFileSystem(log=l, signature='MINIO', config=cfg)
  ttt.upload('f1','dir/obj', bucket_name='aaa')



import traceback
import pandas as pd

from copy import deepcopy
from naeural_core import constants as ct
from time import time
from collections import OrderedDict

from naeural_core.manager import Manager
from naeural_core import Logger
from naeural_core.local_libraries import _ConfigHandlerMixin

DEFAULT_DISALLOWED_URL_DUPLICATES = ['VIDEOSTREAM']
WARNING_REDISPLAY_TIME = 30

class CaptureManager(Manager, _ConfigHandlerMixin):
  def __init__(self, log : Logger, shmem, owner, environment_variables=None, **kwargs):
    self.__name__ = ct.CAPTURE_MANAGER
    shmem[ct.CAPTURE_MANAGER] = {
      ct.NR_CAPTURES : 0, 
      }
    self.shmem = shmem

    self.owner = owner
    self._environment_variables = environment_variables or {}
    self._dct_config_streams = None
    self._dct_captures = None
    self._dct_killed_captures = None
    self._last_gather_summary = None
    self.is_data_available = False
    
    self.data_flow_stopped = False    
    self._data_flow_skip_last_log = 0
    self._nr_data_flow_skips = 0
    self._data_flow_stop_last_log = 0
    self._nr_data_flow_stops = 0


    super(CaptureManager, self).__init__(log=log, prefix_log='[{}]'.format(ct.CAPTURE_MANAGER), **kwargs)
    return
  
  def startup(self):
    super().startup()
    self._dct_captures = self._dct_subalterns
    self._dct_killed_captures = {}
    return

  def update_streams(self, dct_config_streams):
    if self._dct_config_streams != dct_config_streams:
      self.owner.set_loop_stage('4.collect.update_streams.start')
      self._dct_config_streams = dct_config_streams
      self.config_data = self._dct_config_streams
      self.owner.set_loop_stage('4.collect.update_streams._check_captures')
      current_captures = self._check_captures()
      self.owner.set_loop_stage('4.collect.update_streams._deallocate_unused_captures')
      self._deallocate_unused_captures(current_captures)
    #endif
    return
  
  def _maybe_log_dataflow_errors(self, msg=''):
    msg1, msg2 = None, None
    plugins = self.owner.get_overloaded_business_plugins()
    show1 = True #((time() - self._data_flow_skip_last_log) > (WARNING_REDISPLAY_TIME * 1.3)) and len(plugins) > 0
    show2 = True #(time() - self._data_flow_stop_last_log) > WARNING_REDISPLAY_TIME
    if msg == '':
      msg = 'DataFlow errors:'
    if (show1 or show2) and msg:
      self.P(msg, color='r')
    
    if show1:
      self._data_flow_skip_last_log = time()
      msg1 = "  {} skippings for data capture due to the following delayed plugins: {}".format(
        self._nr_data_flow_skips,
        plugins
      )
      self.P(msg1, color='r')
      
    if show2:
      self._data_flow_stop_last_log = time()
      msg2 = "  {} data acquisition stop/restarts so far due to business plugins issues".format(
        self._nr_data_flow_stops,
      )
      self.P(msg2, color='r')
      
    if msg1 is not None or msg2 is not None:
      self._create_notification(
        notif=ct.STATUS_TYPE.STATUS_ABNORMAL_FUNCTIONING,
        msg="DCT Issues: delays in business plugins may generate data loss. Please check configs",
        info="M1:{} M2:{}".format(msg1, msg2),
        displayed=False, # force display
      )
    return
      
    
  
  def _needs_skip_acquisition(self):
    result = False
    if self.owner.any_overloaded_business_plugins:
      result = True
      self._nr_data_flow_skips += 1
    return result
        
    

  def _check_captures(self):
    current_captures = []
    self.shmem[ct.CAPTURE_MANAGER][ct.NR_CAPTURES] = len(self._dct_config_streams)
    streams = list(self._dct_config_streams.keys())
    for s in streams:
      config_stream = self._dct_config_streams[s]
      if config_stream[ct.TYPE].lower() == 'void':
        # void streams are not controlled by capture manager, as they do not provide data.
        continue

      key = config_stream[ct.NAME]
      current_captures.append(key)

      if key not in self._dct_captures:        
        # new capture
        self.owner.set_loop_stage('4.collect.update_streams._check_captures.start_capture.' + key)
        ok_capture = self.start_capture(config_stream)
        if ok_capture:
          self.log.P("Start capture '{}' succedeed.".format(key), color='g')
        else:
          self.P("Start capture '{}' failed.".format(key), color='r')
      else:
        # maybe config it
        capture = self._dct_captures[key]
        self.owner.set_loop_stage('4.collect.update_streams._check_captures.maybe_update_config.' + key)
        capture.maybe_update_config(config_stream)
      #endif
    #endfor    
    return current_captures

  def _deallocate_unused_captures(self, current_captures):
    keys = list(self._dct_captures.keys())
    for k in keys:
      if k not in current_captures:        
        self.stop_capture(k)
    return
  
  def _check_url_duplicate(self, cfg_url, cfg_type, cfg_name):
    if cfg_url is None:
      return 
    disallowed_types = self._environment_variables.get('DISALLOWED_URL_DUPLICATES', [])
    disallowed_types = [x.upper() for x in disallowed_types]
    for default_disallowed_type in DEFAULT_DISALLOWED_URL_DUPLICATES:
      if default_disallowed_type not in disallowed_types:
        disallowed_types.append(default_disallowed_type.upper())
    caps = [x for n,x in self._dct_captures.items() if n != cfg_name and x is not None]
    target_caps = [(cap.cfg_name, cap.cfg_type) for cap in caps if cap.cfg_type.upper() in disallowed_types]
    if cfg_type.upper() in disallowed_types:
      self.P("Checking URL duplicate for '{}' ({}) against: {}".format(cfg_name, cfg_type, target_caps))
    else:
      return
    for cap in caps:
      found_cap = cap.cfg_name
      found_url = cap.cfg_url
      found_type = cap.cfg_type
      
      if (cfg_url == found_url) and (cfg_type == found_type) and cfg_type.upper() in disallowed_types:
        msg = "Data capture '{}' ({}) has duplicate URL '{}' of existing capture '{}' ({}).".format(
          cfg_name, cfg_type, cfg_url, 
          found_cap, found_type,
        )
        self.P(msg, color='r')
        self._create_notification(
          notif=ct.STATUS_TYPE.STATUS_EXCEPTION,
          notif_code=ct.NOTIFICATION_CODES.PIPELINE_FAILED,
          msg=msg,
          stream_name=cfg_name,
          info=msg + ' Recommendation is to use "TYPE" : "Metastream" and "COLLECTED_STREAMS" : ["{}"] within {}'.format(
          found_cap, cfg_name),
          duplicate_url=found_url,
          duplicate_stream_name=found_cap,
          proposed_solution={"COLLECTED_STREAMS" : [found_cap]},
          displayed=True,
        )
        raise ValueError(msg)
      #endif
    #endfor
    return


  def start_capture(self, config):
    stream_name = config[ct.NAME]
    plugin_signature = config[ct.TYPE]
    _cls, _cls_config = self._get_plugin_class(plugin_signature)

    if _cls is None:
      capture = None
      msg = "Error loading capture plugin '{}'".format(plugin_signature)
      self.P(msg + " on stream {}".format(stream_name), color='r')
      self._create_notification(
        notif=ct.STATUS_TYPE.STATUS_EXCEPTION,
        notif_code=ct.NOTIFICATION_CODES.PIPELINE_FAILED,
        msg=msg,
        session_id=config.get(ct.PAYLOAD_DATA.SESSION_ID),
        initiator_id=config.get(ct.PAYLOAD_DATA.INITIATOR_ID),
        stream_name=stream_name,
        info="No code/script defined for capture plugin '{}' in {}".format(plugin_signature, ct.PLUGIN_SEARCH.LOC_DATA_ACQUISITION_PLUGINS),
        ct=ct,
      )
    else:
      try:
        self.P("Attempting to start data capture <{}:{}> ...".format(
          plugin_signature, stream_name), color='m'
        )
        ## pre-instance checks
        ### check if this capture uses a alrady in use URL and raise exception
        cfg_url = config.get(ct.URL, None)
        self._check_url_duplicate(cfg_url=cfg_url, cfg_type=plugin_signature, cfg_name=stream_name)
        ### end check duplicate URL
        ## end pre-instance checks
        capture = _cls(
          log=self.log,
          default_config=_cls_config,
          upstream_config=config,
          environment_variables=self._environment_variables,
          shmem=self.shmem,
          fn_loop_stage_callback=self.owner.set_loop_stage,
          signature=plugin_signature
        )
        if capture.cfg_is_thread:
          capture.start()
          msg = "Started capture thread for pipeline {} (url:{} live:{} recon:{})".format(stream_name, capture.cfg_url, capture.cfg_live_feed, capture.is_reconnectable)
        else:
          msg = "Started META capture: {} (collected: {})".format(stream_name, capture.cfg_collected_streams)
        #endif
        self.P(msg)
        ### IMPORTANT: the following message signals that the pipeline can be used
        self._create_notification(
          notif=ct.STATUS_TYPE.STATUS_NORMAL,
          notif_code=ct.NOTIFICATION_CODES.PIPELINE_OK,
          msg=msg,
          stream_name=stream_name,
          session_id=config.get(ct.PAYLOAD_DATA.SESSION_ID),
          initiator_id=config.get(ct.PAYLOAD_DATA.INITIATOR_ID),
          autocomplete_info=True,
          displayed=True,
        )        
      except Exception as exc:
        # capture is set to None and will trigger archiving from `get_finished_streams` via orchestrator 
        capture = None 
        msg = "Exception '{}' in capture plugin `{}` init for '{}': {}\n{}".format(
          exc, 
          plugin_signature, 
          stream_name,
          self.log.get_error_info(),
          traceback.format_exc(),
        )
        self.P(msg, color='r')
        self._create_notification(
          notif=ct.STATUS_TYPE.STATUS_EXCEPTION,
          notif_code=ct.NOTIFICATION_CODES.PIPELINE_FAILED,
          msg=msg,
          stream_name=stream_name,          
          session_id=config.get(ct.PAYLOAD_DATA.SESSION_ID),
          initiator_id=config.get(ct.PAYLOAD_DATA.INITIATOR_ID),
          autocomplete_info=True,
          displayed=True,
        )
      #end try-except
    #endif

    self._dct_captures[stream_name] = capture
    ok_capture = capture is not None
    return ok_capture

  def _get_plugin_class(self, signature):
    _module_name, _class_name, _class_def, _class_config = self._get_module_name_and_class(
      locations=ct.PLUGIN_SEARCH.LOC_DATA_ACQUISITION_PLUGINS,
      name=signature,
      suffix=ct.PLUGIN_SEARCH.SUFFIX_DATA_ACQUISITION_PLUGINS,
      safe_locations=ct.PLUGIN_SEARCH.SAFE_LOC_DATA_ACQUISITION_PLUGINS,
      safe_imports=ct.PLUGIN_SEARCH.SAFE_LOC_DATA_ACQUISITION_IMPORTS,
      safety_check=True, 
    )
    return _class_def, _class_config


  def stop_capture(self, key, shutdown=False):
    capture = self._dct_captures.pop(key)
    if capture is None:
      self.P("  Capture '{}' already stopped!".format(key), color='r')      
    else:
      cap_type = capture.cfg_type
      self.P("  Stopping capture/pipeline '{}'  ({})...".format(key, cap_type), color='y')
      
      if capture is None:
        return
      self.owner.set_loop_stage('4.collect.capture_stop_{}'.format(key))
      jointime = 0.1 if shutdown else 10 
      capture.stop(join_time=jointime)
      self._dct_killed_captures[key] = None
      self.P("  Stopped capture '{}'".format(key), color='y')
    #endif capture already stopped or not
    return


  def close(self):
    self.P("Stopping all DCTs...", color='y')
    self.stop_captures(shutdown=True)
    self.P("Dope stopping all DCTs...", color='y')
    return


  def stop_captures(self, shutdown=False):
    keys = list(self._dct_captures.keys())
    for k in keys:
      self.stop_capture(k, shutdown=shutdown)
    return


  def get_captured_data(self, key):
    """
    This handles only a DCT based on `key` that is not a metastream. Metastreams
    are handled in `get_all_captured_data` post the call of this function

    Parameters
    ----------
    key : str
      name of the DCT.

    Returns
    -------
    has_data : bool
      Data or no data bool.
      
    captured_data : dict
      the captured data.

    """

    has_data = False
    captured_data = None
    if key in self._dct_captures:
      capture = self._dct_captures[key]
      if capture is not None and capture.cfg_is_thread and capture.has_data:
        self.owner.set_loop_stage('4.collect.get_all_cap.get_captured_data.' + key)
        captured_data = capture.get_data_capture()
        has_data = True
      #endif
    #endif
    return has_data, captured_data

  def _aggregate_captured_data_in_metastreams(self, dct_captured_data):
    """
    This function populates meta-streams with data from normal DCTs
    """
    self.log.start_timer('aggregate_meta')
    for key in self._dct_captures:
      capture = self._dct_captures[key]
      if capture is not None and not capture.cfg_is_thread:
        self.log.start_timer('prepare_meta_capture')
        # is_thread will return True for normal DCT threads while form meta 
        # streams False will return (as it will be simple DataCapture instanaces)
        inputs = []
        all_collected_streams = capture.cfg_collected_streams
        lst_datas = []
        call_metastream_post_process = True
        for collected_stream in all_collected_streams:
          ### if any of the collected streams is not captured, then the meta-stream can't feed data because errors may occur (unless it is not configured to be loose: "IS_LOOSE")
          if collected_stream not in dct_captured_data:
            if capture.cfg_is_loose:
              # default behaviour of meta-streams is to accept partial data
              continue
            else:
              call_metastream_post_process = False
              break
          #endif

          crt_inputs = dct_captured_data[collected_stream]['INPUTS']
          self.log.start_timer('prepare_meta_capture_data')
          for dct_inp in crt_inputs:
            # we take each input and deep-copy it
            dct_res = deepcopy(dct_inp)
            # now we add original metadata & other info
            dct_res['METADATA'] = {
              **dct_res['METADATA'],
              **dct_captured_data[collected_stream]['STREAM_METADATA'],
              **{'SOURCE_STREAM_NAME' : dct_captured_data[collected_stream]['STREAM_NAME']}
            }
            lst_datas.append(dct_res)
          self.log.stop_timer('prepare_meta_capture_data')
        # endfor each collected stream (if any)
        if call_metastream_post_process:
          """
          If we do not hit the break at line 359, we execute the post_process_inputs.
          This means that we found all our captures with data.
          """
          if hasattr(capture, 'post_process_inputs'):
            # we post-process original inputs if required
            self.log.start_timer('prepare_meta_capture_post_proc')
            lst_metastream_inputs = capture.post_process_inputs(lst_datas)
            self.log.stop_timer('prepare_meta_capture_post_proc')
            # finally we add the inputs to the list of inputs
            inputs += lst_metastream_inputs
        # endif call_metastream_post_process
        if len(inputs) > 0:
          # finally we add the date in the captured data dict for downstream consumption
          dct_captured_data[key] = capture.data_template(inputs=inputs, metadata={
            "cap_time":inputs[-1]
              .get('METADATA', {})
              .get('cap_time', 0)
          })
        self.log.stop_timer('prepare_meta_capture')
      #endif is meta-stream
    #endfor
    self.log.stop_timer('aggregate_meta')
    return dct_captured_data

  def get_all_captured_data(self):
    self.log.start_timer('get_all_captures_full')
    
    self.log.start_timer('get_all_captures_stage1')
    self.owner.set_loop_stage('4.collect.get_all_cap.start')
    dct_captured_data = {}
    needs_skip = self._needs_skip_acquisition()
    self.log.stop_timer('get_all_captures_stage1')
    
    if needs_skip:
      self.log.start_timer('get_all_captures_skip')
      self.owner.set_loop_stage('4.collect.get_all_cap.data_flow_stopped')
      if not self.data_flow_stopped:
        self._nr_data_flow_stops += 1
        self._maybe_log_dataflow_errors(msg="Data flow STOPPED.")
      self.data_flow_stopped = True
      self.log.stop_timer('get_all_captures_skip')
    else:
      self.log.start_timer('get_all_captures_stage2')
      self.owner.set_loop_stage('4.collect.get_all_cap.data_flow_working')
      if self.data_flow_stopped:
        self._maybe_log_dataflow_errors(msg="Data flow RESTARTED.")        
      self.data_flow_stopped = False      
      self.owner.set_loop_stage('4.collect.get_all_cap.get_captured_data')
      self.log.stop_timer('get_all_captures_stage2')

      self.log.start_timer('get_all_captures_stage3')      
      lst_avail = [k for k in self._dct_captures if self._dct_captures[k] is not None]
      n = len(lst_avail)
      str_stage = 'get_captured_n{}'.format(n)
      self.log.start_timer(str_stage)
      self.owner.set_loop_stage('4.collect.get_all_cap.get_captured_data.' + str_stage)
      dct_captured_data_raw = {k: self.get_captured_data(k) for k in lst_avail}
      self.log.stop_timer(str_stage)
      
      self.owner.set_loop_stage('4.collect.get_all_cap.dct_captured_data_raw')
      dct_captured_data = {
        name: captured_data for name, (has_data, captured_data) in dct_captured_data_raw.items() if has_data
      }      
      self.log.stop_timer('get_all_captures_stage3')   
         
      self.log.start_timer('get_all_captures_stage4')
      # aggredate date where we have streams
      self.owner.set_loop_stage('4.collect.get_all_cap.aggregate_metastreams')
      dct_captured_data = self._aggregate_captured_data_in_metastreams(dct_captured_data)
      self.log.stop_timer('get_all_captures_stage4')

      # TODO(S+AID): check all `dct_captured_data` images and raise warning if images are not `uint8`!

      self.log.start_timer('get_all_captures_stage_status')
      display_status_each = self.log.config_data.get(ct.CAPTURE_STATS_DISPLAY,  ct.CAPTURE_STATS_DISPLAY_DEFAULT)
      if (self._last_gather_summary is None) or (time() - self._last_gather_summary > display_status_each):
        self._last_gather_summary = time()
        self.owner.set_loop_stage('4.collect.get_all_cap.get_captures_status')
        _ = self.get_captures_status(display=True)
      #endif display stats
      
      if len(dct_captured_data) > 0: 
        # this was "unrolled" for debug purposes, so please do not `self.is_data_available = len(dct_captured_data) > 0`
        self.is_data_available  = True
      else:
        self.is_data_available = False
      self.log.stop_timer('get_all_captures_stage_status')
    #endif needs skip
    self.log.stop_timer('get_all_captures_full')
    return dct_captured_data
  
  
  def get_captures_status(self, display=False, as_dict=True):
    name_maxlen = 15
    n_issues = 0
    info_text = ""
    alerts = {}
    issues = ""
    dct_cap_status = OrderedDict()
    issue_caps = []
    config_startup = self.shmem['config_startup'] 
    eeid = config_startup.get(ct.CONFIG_STARTUP_v2.K_EE_ID, '')[:ct.EE_ALIAS_MAX_SIZE]
    title = "Current active captures on '{}' (only DCT captures shown with no metastreams):".format(eeid)
    dct_msg = OrderedDict()
    # TODO: generalize the stat_cols keys so that they can be adapted to any type of capture
    #       and not only to the current implementation of ffmpeg
    stat_cols = ["Dup", "Drop", "Dec", "Cor", "Miss", "Con", "Delay", ] # this should be a `STAT_COLS` generalized constant with all needed keys 
    all_cols = ["Status" , "Name", "IdlCap", "IdlSnd", "LPS", "CFG/TGT", "DDPS", "Fails"]
    all_cols += stat_cols
    # all_cols += ["Running stats"]
    for col in all_cols:
      dct_msg[col] = []
    now_str = self.log.time_to_str()
    # process all DCTs
    cap_names = list(self._dct_captures.keys())
    maxlen = max([0] + [len(x) for x in cap_names])
    for key in cap_names:
      capture = self._dct_captures[key]
      if capture is None:
        continue
      alerts[key] = capture.is_idle_alert
      if capture.cfg_is_thread:
        dct_cap_status[key] = {
          'TYPE'        : capture.cfg_type,
          'FLOW'        : 'live' if capture.cfg_live_feed else 'buff',
          'IDLE'        : round(capture.time_since_last_input,1),
          'IDLE_ALERT'  : capture.is_idle_alert,
          'DPS'         : round(capture.actual_dps,3),
          'CFG_DPS'     : capture.cfg_cap_resolution,
          'TGT_DPS'     : capture.cap_resolution,
          'RUNSTATS'    : capture.get_runstats(),
          'PLUGINSTATS' : capture.get_plugin_specific_stats(),
          'COLLECTING'  : None,
          'FAILS'       : capture.nr_connection_issues,
          'NOW'         : now_str,
          'GET_DPS'     : capture.generate_resolution,
        }
        dct_msg['Status'].append(dct_cap_status[key]['FLOW'])
        dct_msg['Name'].append(key[:name_maxlen])
        dct_msg['IdlCap'].append(dct_cap_status[key]['IDLE'])
        dct_msg['IdlSnd'].append(round(capture.time_since_last_data,1))
        dct_msg['LPS'].append(dct_cap_status[key]['DPS'])
        dct_msg['CFG/TGT'].append("{}/{}".format(dct_cap_status[key]['CFG_DPS'], dct_cap_status[key]['TGT_DPS']))
        dct_msg['DDPS'].append(dct_cap_status[key]['GET_DPS'])
        dct_msg['Fails'].append(dct_cap_status[key]['FAILS'])
        # dct_msg['Running stats'].append(dct_cap_status[key]['RUNSTATS'])      
        for col in stat_cols:
          dct_msg[col].append(dct_cap_status[key]['PLUGINSTATS'].get(col.lower(), -1))

        if display and alerts[key]: # limit the number of alerts by counting them only at display time
          n_issues += 1
          issues += '\n {} idle: {}'.format(key, dct_cap_status[key]["IDLE"])
          issue_caps.append(key)
        #endif count alerts
      #endif is working thread 
    #endfor
    # process all meta-streams
    for key in cap_names:
      capture = self._dct_captures[key]
      if capture is None:
        continue
      alerts[key] = capture.is_idle_alert
      if not capture.cfg_is_thread:
        dct_cap_status[key] = {
          'TYPE'        : capture.cfg_type,
          'FLOW'        : 'meta',
          'IDLE'        : round(capture.time_since_last_data,1),
          'DPS'         : -1,
          'CFG_DPS'     : -1,
          'TGT_DPS'     : -1,
          'RUNSTATS'    : -1,
          'PLUGINSTATS' : capture.get_plugin_specific_stats(),
          'COLLECTING'  : capture.cfg_collected_streams,
          'FAILS'       : -1,
          'NOW'         : now_str,
        }

        dct_msg['Status'].append("[meta]")
        dct_msg['Name'].append(key[:name_maxlen])
        dct_msg['IdlCap'].append(None)
        dct_msg['IdlSnd'].append(round(capture.time_since_last_data,1))
        dct_msg['LPS'].append(None)
        dct_msg['CFG/TGT'].append(None)
        dct_msg['DDPS'].append(None)
        dct_msg['Fails'].append(None)
        # dct_msg['Running stats'].append(capture.cfg_collected_streams)      
        for col in stat_cols:
          dct_msg[col].append(dct_cap_status[key]['PLUGINSTATS'].get(col.lower(), -1))
        
        if display and alerts[key]: # limit the number of alerts by counting them only at display time
          n_issues += 1
          issues += '\n {} idle: {}'.format(key, dct_cap_status[key]["IDLE"])
          issue_caps.append(key)
        #endif count alerts
      #endif is working thread 
    #endfor
    
    is_alert = any(alerts.values())
    
    

    if display:
      if len(cap_names) == 0:
        info_text = title + "\n  -------  No captures are running at this moment -------"
      else:
        all_cap_timer = self.log._format_timer(
          key='get_all_captures_full',
          section=self.log.default_timers_section,
          was_recently_seen=True,
        )
        prec = pd.get_option('display.float_format')
        _format = '{:.1f}'
        pd.set_option('display.float_format', lambda x: _format.format(x))        
        info_text = "{}\n{}\n----------------------------------------\nTimings: {})".format(
          title, pd.DataFrame(dct_msg),
          all_cap_timer.lstrip()
        )
        pd.set_option("display.float_format", prec)
      #endif
      self.P(info_text, color='r' if is_alert else 'g')
    # endif display
    
    if n_issues > 0:
      msg = "Data acquisition failure. One or more captures idle time over accepted thresholds: {}".format(issue_caps)
      info = msg + '\n' + issues
      self._create_notification(
        notif=ct.STATUS_TYPE.STATUS_ABNORMAL_FUNCTIONING,
        msg=msg,
        info=info,
        errors=issue_caps,
      )
    if as_dict:
      return dct_cap_status
    else:
      return info_text


  def get_finished_streams(self):
    """
    This function returns stream that can be deleted due to:
      - object is None in capture dict (maybe it was `None` since its creation)
      or
      - can_delete (stream is NOT reconnectable and NOT keep-alive)
      or
      - stream has a limited life that has elapsed
    """
    lst = [
      key for key,obj in self._dct_captures.items() 
      if (obj is None or (obj.can_delete or obj.alive_until_passed))
    ]
    return lst

  def get_plugin_default_config(self, signature):
    _module_name, _class_name, _cls_def, _config_dict = self._get_module_name_and_class(
      locations=ct.PLUGIN_SEARCH.LOC_DATA_ACQUISITION_PLUGINS,
      name=signature,
      verbose=0,
      suffix=ct.PLUGIN_SEARCH.SUFFIX_DATA_ACQUISITION_PLUGINS,
      safe_locations=ct.PLUGIN_SEARCH.SAFE_LOC_DATA_ACQUISITION_PLUGINS,
      safe_imports=ct.PLUGIN_SEARCH.SAFE_LOC_DATA_ACQUISITION_IMPORTS,
      safety_check=True, # TODO: should we do this?
    )

    return _config_dict



import tensorflow as tf
import tensorflow.keras.backend as K
from functools import partial


def quantile_loss(y_true, y_pred, Q):
  """
  I = y_pred <= y_true
  loss = Q * (y_true - y_pred) * I + (1 - Q) * (y_pred - y_true) * (1 - I)  
  """
  tf_res = y_true - y_pred
  tf_max = K.maximum(Q * tf_res, (Q-1) * tf_res)
  tf_loss = K.mean(tf_max, axis=-1)
  return tf_loss


def quantile_loss_05(y_true, y_pred):
  return quantile_loss(y_true,y_pred, Q=0.05)


def quantile_loss_95(y_true, y_pred):
  return quantile_loss(y_true,y_pred, Q=0.95)


def focal_loss_softmax(y_true, y_pred, gamma=2.0, alpha=0.25):
  """
  this is the fixed params focal loss for softmax outputs
  
  paper: loss = -alpha*((1-p)^gamma)*log(p) (https://arxiv.org/pdf/1708.02002.pdf)
  """
  
  eps = K.epsilon()
  
  y_pred = K.clip(y_pred, eps, 1 - eps)
  xentopy = - y_true * K.log(y_pred)
  focal_weights = alpha * y_true * K.pow((1 - y_pred), gamma)
  _loss = focal_weights * xentopy
  loss = K.sum(_loss, axis=-1)
  return loss


def focal_loss_sigmoid(y_true, y_pred, gamma=2.0, alpha=0.25):
  """
  this is the fixed params focal loss for binary xentropy

  OBSERVATION: alpha MUST BE inverse-proportional to the frequency of the class. So if
  class 1 has 25% presence then alpha is 0.25
  
  please see also https://github.com/vandit15/Class-balanced-loss-pytorch  
  
    Formula: https://arxiv.org/pdf/1708.02002.pdf
        loss = -alpha_t*((1-p_t)^gamma)*log(p_t)
        
        p_t = y_pred, if y_true = 1
        p_t = 1-y_pred, otherwise
        
        alpha_t = alpha, if y_true=1
        alpha_t = 1-alpha, otherwise
        
  def focal_loss_fixed(y_true, y_pred):
    pt_1 = tf.where(tf.equal(y_true, 1), y_pred, tf.ones_like(y_pred))
    pt_0 = tf.where(tf.equal(y_true, 0), y_pred, tf.zeros_like(y_pred))

    pt_1 = K.clip(pt_1, 1e-3, .999)
    pt_0 = K.clip(pt_0, 1e-3, .999)

    return -K.sum(alpha * K.pow(1. - pt_1, gamma) * K.log(pt_1))-K.sum((1-alpha) * K.pow( pt_0, gamma) * K.log(1. - pt_0))        

  """
  
  eps = K.epsilon()  
  y_pred = K.clip(y_pred, eps, 1 - eps)
  # Calculate p_t
  p_t = tf.where(K.equal(y_true, 1), y_pred, 1-y_pred)
  # Calculate alpha_t
  alpha_factor = K.ones_like(y_true) * alpha
  alpha_t = tf.where(K.equal(y_true, 1), alpha_factor, 1-alpha_factor)
  # Calculate cross entropy
  cross_entropy = -K.log(p_t)
  weight = alpha_t * K.pow((1-p_t), gamma)
  # Calculate focal loss
  loss = weight * cross_entropy
  # Sum the losses in mini_batch
  loss = K.sum(loss, axis=1)
  return loss

  

def get_focal_loss_softmax(gamma=2.0, alpha=0.25):
  return partial(focal_loss_softmax, gamma=gamma, alpha=alpha)

def get_focal_loss_sigmoid(gamma=2.0, alpha=0.25):
  return partial(focal_loss_sigmoid, gamma=gamma, alpha=alpha)




def huber_loss(y_true, y_pred, d=1.0):
  """
  0.5 * x^2                  if |x| <= d
  0.5 * d^2 + d * (|x| - d)  if |x| > d
  """
  tf_res = y_pred - y_true
  cond1 = K.cast((K.abs(tf_res) <= d), tf.float32)
  cond2 = K.cast((K.abs(tf_res) > d), tf.float32)
  tf_batch_loss1 = cond1 * (0.5 * (tf_res**2))
  tf_batch_loss2 = cond2 * (d * (K.abs(tf_res) - 0.5 * d))
  tf_batch_loss = tf_batch_loss1 + tf_batch_loss2
  tf_loss = K.mean(tf_batch_loss)
  return tf_loss

def K_huber_loss(y_true, y_pred, d=1.0):
  return huber_loss(y_true=y_true, y_pred=y_pred, d=d)


def contrastive_loss(y_true, y_pred, margin=1):
  tf_sqpred = K.square(y_pred)
  tf_margin = K.square(K.maximum(margin - y_pred, 0))
  tf_mean = K.mean(y_true * tf_sqpred + (1 - y_true) * tf_margin)
  return tf_mean


def MAD_loss(y_true, y_pred):
  tf_res = K.abs(y_true - y_pred)
  tf_mad = tf_res / (K.abs(y_pred) + K.epsilon())
  return K.mean(tf_mad)


def MADSQ_loss(y_true, y_pred):
  tf_res_sq = K.pow(y_true - y_pred, 2)
  tf_mad_sq = tf_res_sq / (K.pow(y_pred, 2) + K.epsilon())
  return K.mean(tf_mad_sq)

def triplet_loss_variable_margin(y_pred):
  anchor = y_pred[0]
  positive = y_pred[1]
  negative = y_pred[2]
  beta = y_pred[3]

  similar_dist = K.sum(K.square(anchor - positive), axis=1)
  diff_dist = K.sum(K.square(anchor - negative), axis=1)
  dist = K.expand_dims(similar_dist - diff_dist)
  loss = K.maximum(dist + beta, 0.0)
  return loss

def triplet_loss(y_pred, beta=0.5,):

  anchor = y_pred[0]
  positive = y_pred[1]
  negative = y_pred[2]

  similar_dist = K.sum(K.square(anchor - positive), axis=1)
  diff_dist = K.sum(K.square(anchor - negative), axis=1)
  loss = K.maximum(similar_dist - diff_dist + beta, 0.0)
  loss = K.expand_dims(loss)
  return loss


def identity_loss(y_true, y_pred):
  return K.mean(y_pred - 0 * y_true)


def quad_loss(y_pred, beta1, beta2, ):
  
  anchor = y_pred[0]
  positive = y_pred[1]
  negative = y_pred[2]
  negative2 = y_pred[3]

  similar_dist = K.sum(K.square(anchor - positive), axis=1)
  diff_dist1 = K.sum(K.square(anchor - negative), axis=1)
  diff_dist2 = K.sum(K.square(negative - negative2), axis=1)
  basic_loss_1 = similar_dist - diff_dist1 + beta1
  basic_loss_2 = similar_dist - diff_dist2 + beta2
  # now finally max operator but NO summing as this is a layer not a
  # actual loss function
  loss_1 = K.maximum(basic_loss_1, 0.0)
  loss_2 = K.maximum(basic_loss_2, 0.0)
  loss = loss_1 + loss_2
  loss = K.expand_dims(loss)
  return loss






import flask
import numpy as np
import json

from time import sleep

from threading import Lock

from naeural_core import Logger
from ratio1 import BaseDecentrAIObject, _PluginsManagerMixin
from ratio1.logging.logger_mixins.json_serialization_mixin import NPJson

from naeural_core.local_libraries.model_server_v2.request_utils import get_api_request_body, MSCT

from naeural_core.main.ver import __VER__ as __CORE_VER__

class FlaskModelServer(BaseDecentrAIObject, _PluginsManagerMixin):

  app = None

  def __init__(self, log : Logger,
               workers_location,
               worker_name,
               worker_suffix=None,
               microservice_name=None,
               host=None,
               port=None,
               config_endpoint=None,
               execution_path=None,
               verbosity_level=1,
               nr_workers=None
               ):

    """
    Parameters:
    -----------
    log : Logger, mandatory

    workers_location: str, mandatory
      Dotted path of the folder where the business logic of the workers is implemented

    worker_name: str, mandatory
      Which implementation found in `workers_location` is started in this server

    worker_suffix: str, optional
      Which is the suffix of the class name for the started worker
      e.g. if the worker .py file is called 'get_similarity.py', then the name of the class is GetSimilarity<Suffix>.
      If `worker_suffix=Worker`; then the name of the class is GetSimilarityWorker
      The default is None

    host: str, optional
      Host of the gateway
      The default is None ('127.0.0.1')

    port: int, optional
      Port of the gateway
      The default is None (5000)

    config_endpoint: dict, optional
      The configuration of the endpoint:
        * 'NR_WORKERS'
        * other params that will be passed as upstream configuration to the worker
      The default is None ({})

    execution_path: str, optional
      The API rule where the worker logic is executed.
      The default is None ('/analyze')

    verbosity_level: int, optional
      The default is 1

    nr_workers: int, optional
      How many instances of the chosen worker are started.
      The default is None (5)
    """

    self.__version__ =  __CORE_VER__
    self.__workers_location = workers_location
    self.__worker_name = worker_name
    self.__worker_suffix = worker_suffix
    self.__initial_nr_workers = nr_workers or 5
    self.__microservice_name = microservice_name or worker_name
    self.__worker_class_name = None

    self._host = host or '127.0.0.1'
    self._port = port or 5000
    self._execution_path = execution_path or '/analyze'
    self._verbosity_level = verbosity_level
    self._config_endpoint = config_endpoint or {}

    self._lst_workers = []
    self._mask_workers_in_use = []
    self._counter = 0

    self._lock = Lock()
    self._lock_counter = Lock()
    self._paths = None

    super(FlaskModelServer, self).__init__(log=log, prefix_log='[FSKSVR]', maxlen_notifications=1000)
    return

  def startup(self):
    super().startup()
    self._update_nr_workers(self.__initial_nr_workers)
    self._log_banner()

    if not self._execution_path.startswith('/'):
      self._execution_path = '/' + self._execution_path

    self.app = flask.Flask('FlaskModelServer')
    self.app.json_encoder = NPJson
    self.app.add_url_rule(
      rule=self._execution_path,
      endpoint="PluginEndpoint",
      view_func=self._view_func_plugin_endpoint,
      methods=['GET', 'POST', 'OPTIONS']
    )

    self.app.add_url_rule(
      rule=MSCT.RULE_NOTIF,
      endpoint='NotificationsEndpoint',
      view_func=self._view_func_notifications_endpoint,
      methods=['GET', 'POST']
    )

    self.app.add_url_rule(
      rule=MSCT.RULE_UPDATE_WORKERS,
      endpoint='WorkersEndpoint',
      view_func=self._view_func_workers_endpoint,
      methods=['GET', 'POST'],
    )

    self._paths = [self._execution_path, MSCT.RULE_NOTIF, MSCT.RULE_UPDATE_WORKERS]

    self.app.add_url_rule(
      rule=MSCT.RULE_PATHS,
      endpoint='GetPathsEndpoint',
      view_func=self._view_func_get_paths_endpoint,
      methods=['GET', 'POST']
    )

    self.app.run(
      host=self._host,
      port=self._port,
      threaded=True
    )

    return

  def _log_banner(self):
    _logo = "FlaskModelServer v{} '{}' <{} code={}> started on '{}:{}'".format(
      self.__version__, self.__microservice_name, self.__worker_name, self.__worker_class_name,  self._host, self._port
    )

    lead = 5
    _logo = " " * lead + _logo
    s2 = (len(_logo) + lead) * "*"
    self.log.P("")
    self.log.P(s2)
    self.log.P("")
    self.log.P(_logo)
    self.log.P("")
    self.log.P(s2)
    self.log.P("")
    self.log.P("Given full config:\n{}".format(json.dumps(self._config_endpoint, indent=4)))
    return

  def _create_worker(self):
    _module_name, _class_name, _cls_def, _config_dict = self._get_module_name_and_class(
      locations=self.__workers_location,
      name=self.__worker_name,
      suffix=self.__worker_suffix
    )
    
    if _cls_def is None:
      return
    
    self.__worker_class_name = _cls_def.__name__

    worker_id = len(self._mask_workers_in_use)

    worker = _cls_def(
      log=self.log,
      default_config=_config_dict,
      verbosity_level=self._verbosity_level,
      worker_id=worker_id,
      upstream_config=self._config_endpoint
    )

    self._lst_workers.append(worker)
    self._mask_workers_in_use.append(0)
    return

  def _update_nr_workers(self, nr_workers):
    nr_crt_workers = len(self._lst_workers)
    nr_new_workers = nr_workers - nr_crt_workers

    if nr_new_workers > 0:
      str_msg = "Creating {} workers for server '{}'".format(nr_new_workers, self.__worker_name)
      self.P(str_msg, color='g')
      self._create_notification(notif='log', msg=str_msg)
      for _ in range(nr_new_workers):
        self._create_worker()
      str_msg = "Created {} new workers. (were:{}, total:{})".format(nr_new_workers, nr_crt_workers, nr_workers)
      self._create_notification(notif='log', msg=str_msg)
    elif nr_new_workers < 0:
      ###TODO delete only unused
      str_msg = "Should delete {} - not implemented yet".format(-1*nr_new_workers)
      self._create_notification(notif='log', msg=str_msg)
    else:
      str_msg = "Update with no effect, there are already {} workers".format(nr_workers)
      self._create_notification(notif='log', msg=str_msg)
    #endif
    return

  def _mask_worker_locked(self, wid):
    self._mask_workers_in_use[wid] = 1
    return

  def _mask_worker_unlocked(self, wid):
    self._mask_workers_in_use[wid] = 0

  def _find_unlocked_worker(self, counter):
    self._lock.acquire()
    unlocked_workers = np.where(np.array(self._mask_workers_in_use) == 0)[0]
    if unlocked_workers.shape[0] == 0:
      self._lock.release()
      return

    wid = int(np.random.choice(unlocked_workers))
    self._mask_worker_locked(wid)
    self._lock.release()
    return wid

  def _wait_predict(self, data, counter):
    wid = self._find_unlocked_worker(counter)
    if wid is None:
      # All model servers in use. Waiting...

      while wid is None:
        sleep(1)
        wid = self._find_unlocked_worker(counter)
      #endwhile

      # Waiting done.
    #endif

    # now worker is locked...
    worker = self._lst_workers[wid]
    answer = worker.execute(
      inputs=data,
      counter=counter
    )
    self._mask_worker_unlocked(wid)
    return worker, answer, wid

  def _view_func_plugin_endpoint(self):
    self._lock_counter.acquire()
    self._counter += 1
    counter = self._counter
    self._lock_counter.release()
    
    try:
      request = flask.request
      method = request.method
      
      params = get_api_request_body(request=request, log=self.log)
      client = params.get('client', 'unk')
  
      self._create_notification( # TODO: do we really need this notification? get_qa is crazy...
        notif='log',
        msg=(counter, "Received '{}' request {} from client '{}' params: {}".format(
          method, counter, client, params
        ))
      )
      failed_request = False
      err_msg = ''
    except Exception as exc:
      failed_request = True
      err_msg = str(self.log.get_error_info()) # maybe use
      self.P("Request processing generated exception: {}".format(err_msg), color='r')

    worker, wid = None, -1
    if method != 'OPTIONS' and not failed_request:
      worker, answer, wid = self._wait_predict(data=params, counter=counter)
    else:
      answer = {'request_error' : err_msg}

    if answer is None:
      jresponse = flask.jsonify({
        "ERROR": "input json does not contain right info or other error has occured",
        "client": client,
        "call_id": counter,
        "input": params,
        'core_ver' :  __CORE_VER__,
        'time' : self.log.time_to_str(),
      })
    else:
      if isinstance(answer, dict):
        answer['call_id'] = counter
        answer['core_ver'] =  __CORE_VER__
        answer['time'] = self.log.time_to_str()
        if worker is not None:
          answer['signature'] = '{}:{}'.format(worker.__class__.__name__, wid)
        jresponse = flask.jsonify(answer)
      else:
        assert isinstance(answer, str)
        jresponse = flask.make_response(answer)
      #endif
    #endif

    jresponse.headers["Access-Control-Allow-Origin"] = "*"
    jresponse.headers["Access-Control-Allow-Methods"] = "POST, GET, OPTIONS, DELETE"
    jresponse.headers["Access-Control-Allow-Headers"] = "Content-Type"
    return jresponse

  def _view_func_notifications_endpoint(self):
    server_notifs = self.get_notifications()
    workers_notifs = self.log.flatten_2d_list([w.get_notifications() for w in self._lst_workers])

    all_notifs = server_notifs + workers_notifs

    lst_general_notifs = []
    dct_notifs_per_call = {}
    for notif in all_notifs:
      dct = {
        MSCT.NOTIF_NOTIFICATION_TYPE : notif[MSCT.NOTIF_NOTIFICATION_TYPE],
        MSCT.NOTIF_MODULE : notif[MSCT.NOTIF_MODULE],
        MSCT.TIME   : notif[MSCT.NOTIF_TIME]
      }

      if isinstance(notif['NOTIFICATION'], tuple):
        counter, msg = notif['NOTIFICATION']
        counter = str(counter)
        dct['NOTIF'] = msg

        if counter not in dct_notifs_per_call:
          dct_notifs_per_call[counter] = []

        dct_notifs_per_call[counter].append(dct)
      elif isinstance(notif['NOTIFICATION'], str):
        msg = notif['NOTIFICATION']
        dct['NOTIF'] = msg
        lst_general_notifs.append(dct)
      #endif

    #endfor

    jresponse = flask.jsonify({
      'core_ver' :  __CORE_VER__, 
      **{"GENERAL" : lst_general_notifs},
      **dct_notifs_per_call
    })
    return jresponse

  def _view_func_workers_endpoint(self):
    request = flask.request
    params = get_api_request_body(request=request, log=self.log)

    nr_workers = params.get(MSCT.NR_WORKERS, None)
    if nr_workers is None:
      return flask.jsonify({'core_ver' :  __CORE_VER__, 'ERROR' : "Bad input. 'NR_WORKERS' not found"})

    self._update_nr_workers(nr_workers)
    return flask.jsonify({'MESSAGE': 'OK'})

  def _view_func_get_paths_endpoint(self):
    return flask.jsonify({MSCT.PATHS : self._paths})



# YOLOv5 🚀 by Ultralytics, GPL-3.0 license
"""
Experimental modules
"""

import numpy as np
import torch
import torch.nn as nn


import sys
from pathlib import Path
FILE = Path('utils/y5/models/yolo.py').resolve()
ROOT = FILE.parents[1]  # YOLOv5 root directory
if str(ROOT) not in sys.path:
    sys.path.append(str(ROOT))  # add ROOT to PATH
# ROOT = ROOT.relative_to(Path.cwd())  # relative

from models.common import Conv
from y5utils.downloads import attempt_download


class CrossConv(nn.Module):
    # Cross Convolution Downsample
    def __init__(self, c1, c2, k=3, s=1, g=1, e=1.0, shortcut=False):
        # ch_in, ch_out, kernel, stride, groups, expansion, shortcut
        super().__init__()
        c_ = int(c2 * e)  # hidden channels
        self.cv1 = Conv(c1, c_, (1, k), (1, s))
        self.cv2 = Conv(c_, c2, (k, 1), (s, 1), g=g)
        self.add = shortcut and c1 == c2

    def forward(self, x):
        return x + self.cv2(self.cv1(x)) if self.add else self.cv2(self.cv1(x))


class Sum(nn.Module):
    # Weighted sum of 2 or more layers https://arxiv.org/abs/1911.09070
    def __init__(self, n, weight=False):  # n: number of inputs
        super().__init__()
        self.weight = weight  # apply weights boolean
        self.iter = range(n - 1)  # iter object
        if weight:
            self.w = nn.Parameter(-torch.arange(1., n) / 2, requires_grad=True)  # layer weights

    def forward(self, x):
        y = x[0]  # no weight
        if self.weight:
            w = torch.sigmoid(self.w) * 2
            for i in self.iter:
                y = y + x[i + 1] * w[i]
        else:
            for i in self.iter:
                y = y + x[i + 1]
        return y


class MixConv2d(nn.Module):
    # Mixed Depth-wise Conv https://arxiv.org/abs/1907.09595
    def __init__(self, c1, c2, k=(1, 3), s=1, equal_ch=True):
        super().__init__()
        groups = len(k)
        if equal_ch:  # equal c_ per group
            i = torch.linspace(0, groups - 1E-6, c2).floor()  # c2 indices
            c_ = [(i == g).sum() for g in range(groups)]  # intermediate channels
        else:  # equal weight.numel() per group
            b = [c2] + [0] * groups
            a = np.eye(groups + 1, groups, k=-1)
            a -= np.roll(a, 1, axis=1)
            a *= np.array(k) ** 2
            a[0] = 1
            c_ = np.linalg.lstsq(a, b, rcond=None)[0].round()  # solve for equal weight indices, ax = b

        self.m = nn.ModuleList([nn.Conv2d(c1, int(c_[g]), k[g], s, k[g] // 2, bias=False) for g in range(groups)])
        self.bn = nn.BatchNorm2d(c2)
        self.act = nn.LeakyReLU(0.1, inplace=True)

    def forward(self, x):
        return x + self.act(self.bn(torch.cat([m(x) for m in self.m], 1)))


class Ensemble(nn.ModuleList):
    # Ensemble of models
    def __init__(self):
        super().__init__()

    def forward(self, x, augment=False, profile=False, visualize=False):
        y = []
        for module in self:
            y.append(module(x, augment, profile, visualize)[0])
        # y = torch.stack(y).max(0)[0]  # max ensemble
        # y = torch.stack(y).mean(0)  # mean ensemble
        y = torch.cat(y, 1)  # nms ensemble
        return y, None  # inference, train output


def attempt_load(weights, map_location=None, inplace=True, fuse=True):
    from models.yolo import Detect, Model

    # Loads an ensemble of models weights=[a,b,c] or a single model weights=[a] or weights=a
    model = Ensemble()
    for w in weights if isinstance(weights, list) else [weights]:
        _fn = attempt_download(w)
        ckpt = torch.load(_fn, map_location=map_location)  # load
        m = ckpt['ema' if ckpt.get('ema') else 'model']
        m = m.float()
        if fuse:
          m = m.fuse()
        m.eval()          
        model.append(m)  


    # Compatibility updates
    for m in model.modules():
        if type(m) in [nn.Hardswish, nn.LeakyReLU, nn.ReLU, nn.ReLU6, nn.SiLU, Detect, Model]:
            m.inplace = inplace  # pytorch 1.7.0 compatibility
        elif type(m) is Conv:
            m._non_persistent_buffers_set = set()  # pytorch 1.6.0 compatibility

    if len(model) == 1:
        return model[-1]  # return model
    else:
        print(f'Ensemble created with {weights}\n')
        for k in ['names']:
            setattr(model, k, getattr(model[-1], k))
        model.stride = model[torch.argmax(torch.tensor([m.stride.max() for m in model])).int()].stride  # max stride
        return model  # return ensemble



"""
Unified Video Stream with dynamic backend selection

This plugin acts like any other plugin, but it dynamically selects the backend.

This means that the user can create this plugin, specify the desired backend in the config,
as well as the config for the backend, and the plugin will automatically select the correct
backend and initialize it. After this, this plugin will act like the backend plugin.
"""

from ratio1 import _PluginsManagerMixin

from naeural_core import DecentrAIObject
from naeural_core import constants as ct
from naeural_core.data.base import DataCaptureThread

_CONFIG = {
  **DataCaptureThread.CONFIG,

  'OVERWRITE_DEFAULT_PLUGIN_CONFIG': False,
  'DEFAULT_PLUGINS': {
    'VIDEO_LOSS_01': [{
    }],
  },

  'BACKEND': 'ffmpeg',

  'VALIDATION_RULES': {
    **DataCaptureThread.CONFIG['VALIDATION_RULES'],

  },
}


class VideoStreamDataCapture(DecentrAIObject, _PluginsManagerMixin):
  CONFIG = _CONFIG

  def __init__(self, *args, **kwargs):
    super(VideoStreamDataCapture, self).__init__(*args, **kwargs)

    default_config = kwargs.pop('default_config', {})
    upstream_config = kwargs.get('upstream_config', {})
    str_backend = upstream_config.get('BACKEND', default_config.get('BACKEND'))

    if str_backend is None:
      raise ValueError("Backend not specified")

    if not isinstance(str_backend, str):
      raise ValueError("Backend must be a string")

    signature = 'VIDEO_STREAM_' + str_backend.upper()

    _module_name, _class_name, _cls_def, _config_dict = self._get_module_name_and_class(
      locations=ct.PLUGIN_SEARCH.LOC_DATA_ACQUISITION_PLUGINS,
      name=signature,
      suffix=ct.PLUGIN_SEARCH.SUFFIX_DATA_ACQUISITION_PLUGINS,
      search_in_packages=ct.PLUGIN_SEARCH.SEARCH_IN_PACKAGES,
      safe_locations=ct.PLUGIN_SEARCH.SAFE_LOC_DATA_ACQUISITION_PLUGINS,
      safe_imports=ct.PLUGIN_SEARCH.SAFE_LOC_DATA_ACQUISITION_IMPORTS,
      safety_check=True,
    )

    kwargs['default_config'] = _config_dict
    self.backend = _cls_def(*args, **kwargs)
    return

  def __getattr__(self, name):
    # if the backend is not initialized, we return the attribute from this class
    if 'backend' not in self.__dict__:
      if name not in self.__dict__:
        raise AttributeError(f"'{self.__class__.__name__}' object has no attribute '{name}'")
      return self.__dict__[name]
    return getattr(self.backend, name)

  def __setattr__(self, name, value):
    # If the attribute is backend, set it directly on the instance
    if 'backend' not in self.__dict__:
      super().__setattr__(name, value)
    else:
      # Otherwise, set the attribute on the backend
      setattr(self.backend, name, value)



from .base_data import BaseDataCapture
from .base_plugin_dct import DataCaptureThread
from .abstract_map_reduce import AbstractMapReduceDataCapture
from .base_structured import BaseStructuredDataCapture



import numpy as np

from collections import deque
from naeural_core.business.utils import np_vec_iou

class ImmovableObjectTracker(object):
  def __init__(self, object_max_age, iou_threshold, max_history, min_history=0):
    self._immovable_objects = {}
    self._last_track_id_key = 0
    self._object_max_age = object_max_age
    self._max_history = max_history
    self._iou_threshold = iou_threshold
    self._min_history = min_history

  def _increment_age(self):
    for key in self._immovable_objects.keys():
      self._immovable_objects[key]['age'] += 1
    return

  def _remove_old_objects(self):
    lst_to_be_removed = []
    for key in self._immovable_objects.keys():
      if self._immovable_objects[key]['age'] > self._object_max_age:
        lst_to_be_removed.append(key)
      #endif
    #endfor
    for key in lst_to_be_removed:
      self._immovable_objects.pop(key)
    return

  def _update_object(self, object, detection): ##TODO rename
    object['age'] = 0
    object['tlbr_history'].append(detection)

  def _add_object(self, detection):
    self._last_track_id_key += 1
    self._immovable_objects[self._last_track_id_key] = {
      'age': 0,
      'tlbr_history': deque(maxlen=self._max_history)
    }
    self._immovable_objects[self._last_track_id_key]['tlbr_history'].append(detection)

  def _get_object_tlbr(self, object):
    return np.mean([x for x in object['tlbr_history']], axis=0)

  def _get_np_tlbrs(self):
    return list(self._immovable_objects.keys()), np.array([self._get_object_tlbr(x) for x in self._immovable_objects.values()]) ##TODO

  def _update_detections(self, detections):
    iou = None

    np_detections = np.array(detections)
    object_keys, np_objects = self._get_np_tlbrs()

    if np_detections.shape[0] > 0 and np_objects.shape[0] > 0:
      iou, _,_ = np_vec_iou(np_detections, np_objects)

    for i, detection in enumerate(detections):
      if iou is not None and np.max(iou[i,:]) > self._iou_threshold:
        self._update_object(self._immovable_objects[object_keys[np.argmax(iou[i,:])]], detection)
        continue
      #endif

      self._add_object(detection)
      #endif
    #endfor

  def get_objects(self):
    return [
      {'TLBR_POS': self._get_object_tlbr(x)}
      for x in self._immovable_objects.values()
      if len(x['tlbr_history']) > self._min_history
    ]

  def update_objects(self, detections):
    self._increment_age()
    self._update_detections(detections)
    self._remove_old_objects()



from multiprocessing import Pipe, Queue
from time import time, sleep


COMM_QUEUE_SIZE = 5


def comm_generator(engine='queue'):
  comm_server, comm_client = None, None
  if engine.lower() == 'pipe':
    pipe_server, pipe_client = Pipe(True)
    comm_server = CommEngine(pipe_server)
    comm_client = CommEngine(pipe_client)
  elif engine.lower() == 'queue':
    # raise ValueError("Not implemented")
    queue_server, queue_client = Queue(COMM_QUEUE_SIZE), Queue(COMM_QUEUE_SIZE)
    comm_server = CommEngine((queue_server, queue_client))
    comm_client = CommEngine((queue_client, queue_server))
  elif engine.lower() == 'raw_pipe':
    comm_server, comm_client = Pipe(True)
  # endif engine type
  return comm_server, comm_client


class CommEngine():
  def __init__(self, comm_obj, **kwargs):
    self.__comm = comm_obj
    if isinstance(comm_obj, tuple):
      self.__type = comm_obj[0].__class__.__name__.lower()
    else:
      self.__type = comm_obj.__class__.__name__.lower()
    return
  
  def send(self, data):
    if 'pipe' in self.__type:
      self.__comm.send(data)
    elif 'queue' in self.__type:
      self.__comm[0].put(data)
    return
  
  def recv(self):
    res = None
    if 'pipe' in self.__type:
      res = self.__comm.recv()
    elif 'queue' in self.__type:
      res = self.__comm[1].get()
    return res
  
  def poll(self, timeout):
    res = False
    if 'pipe' in self.__type:
      res = self.__comm.poll(timeout)
    elif 'queue' in self.__type:
      start_time = time()
      while time() - start_time < timeout and not res:
        sleep(0.001)
        res = not self.__comm[1].empty()
      # endwhile queue not empty
    return res

  def close(self):
    if 'pipe' in self.__type:
      self.__comm.close()
    elif 'queue' in self.__type:
      self.__comm[0].close()
      self.__comm[1].close()
    # endif
    return



from naeural_core import constants as ct
import cv2

class _VideoConfigMixin(object):

  def __init__(self):
    self._resize_h, self._resize_w = None, None
    super(_VideoConfigMixin, self).__init__()
    return

  def _force_video_stream_config(self):
    if not self.has_cap_resolution_config:
      str_err = "Cannot start {} due to missing capture resolution configuration!".format(self.__class__.__name__)
      self.P(str_err, color='error')
      raise ValueError(str_err)
    
  
  def _maybe_configure_frame_crop_resize(self, height, width):
    ### universal video stream code
    # if we need live resize
    resize_h = self.cfg_stream_config_metadata.get(ct.RESIZE_H, None)
    resize_w = self.cfg_stream_config_metadata.get(ct.RESIZE_W, None)
    if resize_h is not None and resize_h > 0:
      ratio = height / width
      resize_w = resize_h / ratio if resize_w is None else resize_w
      self._resize_h = int(resize_h)
      self._resize_w = int(resize_w)
      self.P("Live resize enabled from {} to {}".format(
        (height, width), (self._resize_h, self._resize_w)
        ))
      self._metadata.resize_hw = self._resize_h, self._resize_w
    # end if we need live resize
    
    # if we need live crop / reframe 
    self._frame_crop = None
    frame_crop = self.cfg_stream_config_metadata.get(
      ct.FRAME_CROP, 
      self.cfg_stream_config_metadata.get(
        ct.CAPTURE_CROP,
        None,
      )
    )
    if frame_crop is not None:
      if self._resize_h is None:          
        # test if not valid crop to resume normal frame
        if (not isinstance(frame_crop, list) 
            or len(frame_crop) != 4 
            or frame_crop[0] > frame_crop[2]
            or frame_crop[1] > frame_crop[3]
            ):
          self.P(" {}: {} invalid. Resuming capture to normal source stream resolution.".format(
            ct.FRAME_CROP, frame_crop))
        else:
          self._frame_crop = frame_crop
          height = self._frame_crop[2] - self._frame_crop[0]
          width = self._frame_crop[3] - self._frame_crop[1]
          self.P("  Enabling cropping from {} to {} (TLBR:{})".format(
            (self._metadata.frame_h, self._metadata.frame_w), (height, width),
            frame_crop,
          ), color='y')
          self._metadata.frame_h = height
          self._metadata.frame_w = width  
      # end frame cropping
    ### end universal video stream code    
    return
  
  def _maybe_resize_crop(self, img):
    if self._resize_h is not None:
      # if RESIZE_H is enabled in STREAM_CONFIG_METADATA then resize live
      self.start_timer('cv2_resize')
      img = cv2.resize(img, dsize=(self._resize_w, self._resize_h))
      self.end_timer('cv2_resize')
    elif self._frame_crop is not None:
      self.start_timer('cv2_crop')
      _top, _left, _bottom, _right = self._frame_crop
      img = img[_top:_bottom, _left:_right]  
      self.end_timer('cv2_crop')
    return img



from naeural_core import Logger  

if __name__ == '__main__':
  
  l = Logger('TST', base_folder='.', app_folder='_cache')
  
  endpoint = 'MINIO_ENDPOINT'
  access_key = 'MINIO_ACCESS_KEY'
  secret_key = 'MINIO_SECRET_KEY'
  bucket_name ="MINIO_BUCKET_NAME"
  minio_object_name = 'minio:y5/coco.txt'
  http_object_name = 'https://www.dropbox.com/s/krzygbdl6qkzf8e/coco.txt?dl=1'
  
  for url, fn in zip([minio_object_name, http_object_name],['from_minio.txt', 'from_http.txt']):    
      down = l.maybe_download_model(
        url=url, 
        model_file=fn,
        endpoint=endpoint,
        secret_key=secret_key,
        bucket_name=bucket_name,
        access_key=access_key,
      )
    
    
    
 


from .alerter_mixin import _AlerterMixin
from .cmdapi import _CmdAPIMixin
from .daily_interval_mixin import _DailyIntervalMixin
from .dataapi import _DataAPIMixin
from .dataset_builder_mixin import _DatasetBuilderMixin
from .deapi import _DeAPIMixin
from .diskapi import _DiskAPIMixin
from .emailer_mixin import _EmailerMixin
from .execution_audit_mixin import _ExecutionAuditMixin
from .interval_aggregation_mixin import _IntervalAggregationMixin
from .limited_data_mixin import _LimitedDataMixin
from .shm_mixin import _ShmMixin
from .threading import _ThreadingAPIMixin
from .upload_mixin import _UploadMixin
from .utilsapi import _GenericUtilsApiMixin
from .working_hours_mixin import _WorkingHoursMixin
from .state_machine_api import _StateMachineAPIMixin


import cv2
import numpy as np

from decentra_vision.draw_utils import DrawUtils
from shapely.geometry import Polygon, LineString
from naeural_core import Logger

if __name__ == '__main__':
  cfg_file = 'main_config.txt'
  log = Logger(lib_name='VB', config_file=cfg_file, max_lines=1000)
  painter = DrawUtils(log=log)

  box_tlbr = [300, 600, 360, 640]
  top, left, bottom, right = box_tlbr
  box_points = [
    [left, top],
    [right, top],
    [right, bottom],
    [left, bottom],
    [left, top],
  ]

  line_points = [
      [187, 50],
      [195, 74],
      [204, 100],
      [220, 123],
      [237, 140],
      [252, 161],
      [274, 176],
      [276, 190],
      [303, 205],
      [320, 223],
      [342, 232],
      [378, 254],
      [408, 265],
      [439, 279],
      [476, 292],
      [515, 305],
      [538, 312],
      [568, 320],
      [591, 322],
      [619, 329],
      [658, 336],
      [691, 342],
      [751, 345],
      [776, 345],
      [802, 344],
      [830, 344],
      [857, 345],
      [884, 342],
      [917, 339],
      [944, 328],
      [969, 319],
      [990, 302],
      [1000, 273],
      [1003, 249],
      [1012, 222],
      [1035, 211],
      [1064, 200],
      [1089, 201]]

  # box = Polygon(box_points)
  # line = LineString(line_points)
  # target = Polygon(target_points)
  # pts_inter = list(box.intersection(target).exterior.coords)
  # pts_inter = np.array(pts_inter).astype(np.int32).tolist()
  # left, top, right, bottom = box.intersection(target).bounds
  img = np.ones((1080, 1920, 3), dtype=np.uint8) * 255

  img = painter.polygon(
    image=img,
    pts=box_points,
    color=(255, 0, 0)
  )
  # img = painter.polygon(
  #   image=img,
  #   pts=target_points,
  #   color=(0, 255, 0)
  #   )
  # img = painter.polygon(
  #   image=img,
  #   pts=pts_inter,
  #   color=(255, 255, 0)
  #   )
  img = painter.polygon(
    image=img,
    pts=line_points,
    color=(255, 255, 0)
  )

  cv2.imshow('test', img)
  cv2.waitKey(0)
  cv2.destroyAllWindows()

  lst_road = []
  dct_points = {(p[0], p[1]): False for p in line_points}

  line = LineString(line_points)
  box_poly = Polygon(box_points)
  intersect = box_poly.intersection(line)
  diff = box_poly.difference(line)

  if box_poly.intersects(line):
    # blur/noblur....

    # vanish line
    for line_point in line_points:
      line_point_x, line_point_y = line_point
      in_box = False
      if line_point_x >= left and \
              line_point_x <= right and \
              line_point_y >= top and \
              line_point_y <= bottom:
        in_box = True
        dct_points[(line_point_x, line_point_y)] = True
  else:
    print('box does not intersect line')

  print(line_points)



import cv2
vidcap = cv2.VideoCapture('/home/work/Downloads/people_walking.mp4')
success,image = vidcap.read()
count = 0
while success:
  if count % 10 == 0:
    cv2.imwrite("/home/work/Downloads/people_walking_frames/frame%s.jpg" % str(count).zfill(4), image)     # save frame as JPEG file      
    success,image = vidcap.read()
    print('Read a new frame: ', success)
  count += 1


import os
import numpy as np
import torch as th
import cv2
import functools
from naeural_core import Logger
from typing import Union, List

def read_image(path_to_img, mode='rgb', max_height=None):
  _modes = ['rgb', 'gray']
  mode = mode.lower()
  if mode == 'rgb' or mode not in _modes:
    np_img = cv2.cvtColor(cv2.imread(path_to_img), cv2.COLOR_BGR2RGB)
  elif mode == 'gray':
    np_img = np.expand_dims(cv2.cvtColor(cv2.imread(path_to_img), cv2.COLOR_BGR2GRAY), 0)

  if max_height is not None and np_img.shape[0] > max_height:
    ratio = np_img.shape[0] / max_height
    new_w = np_img.shape[1] / ratio
    np_img = cv2.resize(np_img, (int(new_w), max_height))
  # endif should resize

  return np_img

def to_device(device_key):
  def decorator(func):
    @functools.wraps(func)
    def wrapper(*args, **kwargs):
      owner = args[0]

      new_args = [owner]
      for inp in args[1:]:
        retransform = False
        if not isinstance(inp, (list, tuple)):
          inp = [inp]
          retransform = True

        new_inp = []
        for _x in inp:
          if isinstance(_x, np.ndarray):
            new_inp.append(th.tensor(_x, device=vars(owner)[device_key]))
          elif _x is not None:
            new_inp.append(_x.to(vars(owner)[device_key]))
          else:
            new_inp.append(_x)

        if retransform:
          new_inp = new_inp[0]

        new_args.append(new_inp)
      #endfor

      res = func(*new_args, **kwargs)
      return res

    return wrapper

  return decorator

def get_shallow_dataset(log : Logger, path_to_dataset : str, extensions : Union[List[str], str] = None):
  if isinstance(extensions, str):
    extensions = [extensions]

  def _get_dirs(path):
    _lst_dirs = os.listdir(path)
    _lst_dirs = list(filter(
      lambda x: os.path.isdir(os.path.join(path, x)),
      _lst_dirs
    ))
    return sorted(_lst_dirs)
  #enddef

  def _filter_files(lst_files):
    lst_files = sorted(list(filter(
      lambda x: not x.startswith('.'),
      lst_files
    )))

    if extensions is None:
      return lst_files

    return list(filter(
      lambda x: any([x.endswith(ext) for ext in extensions]),
      lst_files
    ))
  #enddef

  lst_classes = _get_dirs(path_to_dataset)
  if len(lst_classes) == 0:
    lst_classes = ['']

  dataset_info = {
    'class_to_id' : dict(zip(lst_classes, list(range(len(lst_classes))))),
    'categ_to_id_per_lvl' : [],
    'path_to_idpath' : [],
    'paths' : [],
  }
  dataset_info['categ_to_id_per_lvl'].append(dataset_info['class_to_id'])

  for cls, lbl in dataset_info['class_to_id'].items():
    path_to_cls = os.path.join(path_to_dataset, cls)

    for root, dirs, files in os.walk(path_to_cls):
      relative_path = root.replace(path_to_cls, '').lstrip(os.sep)
      level = 1
      if relative_path != '':
        level = len(relative_path.split(os.sep)) + 1
      good_files = _filter_files(files)
      if len(good_files) == 0:
        # distributions ... all levels handled here
        if len(dataset_info['categ_to_id_per_lvl']) < level+1:
          dataset_info['categ_to_id_per_lvl'].append({})

        crt_registered_categs = list(dataset_info['categ_to_id_per_lvl'][level].keys())
        lst_new_categs = list(set(dirs) - set(crt_registered_categs))
        dct_new_categs = {d: len(crt_registered_categs) + i for i, d in enumerate(lst_new_categs)}
        dataset_info['categ_to_id_per_lvl'][level] = {
          **dataset_info['categ_to_id_per_lvl'][level],
          **dct_new_categs
        }
      else:
        # images
        id_path = [lbl]
        if relative_path != '':
          for lidx, categ in enumerate(relative_path.split(os.sep)):
            id_categ = dataset_info['categ_to_id_per_lvl'][lidx+1][categ]
            id_path.append(id_categ)

        dataset_info['paths'] += list(map(lambda x: os.path.join(root, x), good_files))
        dataset_info['path_to_idpath'] += [id_path for _ in range(len(good_files))]
      #endif
    #endfor - os.walk
  #endfor - cls

  dataset_info['path_to_idpath'] = np.array(dataset_info['path_to_idpath'])
  return dataset_info


import os
import numpy as np
import torch as th
import json




class TestWrapper(th.nn.Module):
  def __init__(self, fn):
    super(TestWrapper, self).__init__()
    cfg = {'config.txt' : ''}
    self.script_model = th.jit.load(fn, _extra_files=cfg)
    self.config = json.loads(cfg['config.txt' ].decode('utf-8'))
    return

  def forward(self, x):
    x = self.script_model(x)
    x = x + 1
    return x


if __name__ == '__main__':
  GENERATE = False

  folder = os.path.split(__file__)[0]
  fn = os.path.join(folder, 'test.ths')
  
  if GENERATE:
    class Tester(th.nn.Module):
      def __init__(self, func):
        super(Tester, self).__init__()
        self.func = func
        return
      
      def forward(self, inputs):
        th_x = self.func(inputs)  
        return th_x
        
    @th.jit.script
    def func(inputs):
      th_x = inputs * 0
      for i in range(inputs.shape[0]):
        x = inputs[i]
        th_x[i] =  x  + (i + 1)
      return th_x    
    
    m = Tester(func)
    t1 = th.tensor(np.ones((2,5)))
    d = {'inputs' : list(t1.shape)}
    ts = th.jit.trace(m, t1, strict=False)
    extra_files = {'config.txt' : json.dumps(d)}
    ts.save(fn, _extra_files=extra_files)
  else:
    
    model = TestWrapper(fn)
    t2 = th.tensor(np.ones((3,5)))
    with th.inference_mode():
      res1 = model(t2)
  